from collections import OrderedDict
from typing import Dict

import chromadb
import pandas as pd
from chromadb import API
from chromadb.config import Settings
from integrations.handlers.chromadb_handler.settings import DEFAULT_EMBEDDINGS_MODEL
from langchain.vectorstores import Chroma
from mindsdb_sql import ASTNode, CreateTable, Insert, Select, Star

from mindsdb.integrations.handlers.chromadb_handler.helpers import (
    documents_to_df,
    extract_collection_name,
    get_metadata_filter,
    load_embeddings_model,
    split_documents,
)
from mindsdb.integrations.libs.base import VectorStoreHandler
from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE
from mindsdb.integrations.libs.response import RESPONSE_TYPE
from mindsdb.integrations.libs.response import HandlerResponse as Response
from mindsdb.integrations.libs.response import HandlerStatusResponse as StatusResponse
from mindsdb.utilities import log

# todo create separate util dir for vectorstore


class ChromaDBHandler(Chroma, VectorStoreHandler):
    """This handler handles connection and execution of the ChromaDB statements."""

    name = "chromadb"

    def __init__(self, name: str, **kwargs):

        self._connection_data = kwargs.get("connection_data")

        self._client_config = self._client_config = {
            "chroma_api_impl": self._connection_data.get("chroma_api_impl"),
            "chroma_server_host": self._connection_data.get("chroma_server_host"),
            "chroma_server_http_port": self._connection_data.get(
                "chroma_server_http_port"
            ),
        }

        self._collection_name = self._connection_data.get(
            "collection_name", "default_collection"
        )
        self._embedding_model_name = self._connection_data.get(
            "embedding_function", DEFAULT_EMBEDDINGS_MODEL
        )

        self._embedding_function = load_embeddings_model(self._embedding_model_name)

        self._persist_directory = self._connection_data.get("persist_directory")
        self._collection_metadata = self._connection_data.get("collection_metadata")

        self._client = chromadb.Client()
        self.is_connected = True

        VectorStoreHandler.__init__(self, name)

        Chroma.__init__(
            self,
            client=self._client,
            client_settings=self._connection_data,
            embedding_function=self._embedding_function,
            collection_name=self._collection_name,
            collection_metadata=self._collection_metadata,
            persist_directory=self._persist_directory,
        )

    def __del__(self):
        if self.is_connected is True:
            self.disconnect()

    def connect(self) -> API:
        """Connect to a ChromaDB database.

        Returns:
            API: The ChromaDB _client.
        """

        if self.is_connected is True:
            return self._client

        try:

            self._client = chromadb.Client(Settings(**self._client_config))
            self.is_connected = True
        except Exception as e:
            log.logger.error(f"Error connecting to ChromaDB client, {e}!")

        return self._client

    def disconnect(self):
        """Close the database connection."""

        if self.is_connected is False:
            return

        self._client = None
        self.is_connected = False

    def check_connection(self):
        """Check the connection to the ChromaDB database."""

        responseCode = StatusResponse(False)
        need_to_close = self.is_connected is False

        try:
            self._client.heartbeat()
            responseCode.success = True
        except Exception as e:
            log.logger.error(f"Error connecting to ChromaDB , {e}!")
            responseCode.error_message = str(e)
        finally:
            if responseCode.success is True and need_to_close:
                self.disconnect()
            if responseCode.success is False and self.is_connected is True:
                self.is_connected = False

        return responseCode

    def filter_query(self, query: ASTNode) -> Dict:
        """Converts WHERE clause to ChromaDB query syntax to filter Chromadb collection.

        Args:
            query (ASTNode): The query to filter.

        Returns:
            Dict: filter for collection.
        """

        where = {}
        if query.where.op == "and":
            for arg in query.where.args:
                if arg.op == "=":

                    if arg.args[0].parts[-1] == "meta_data_filter":
                        # filters on metadata
                        where["meta_data_filter"] = get_metadata_filter(
                            arg.args[1].value
                        )

                    elif arg.args[0].parts[-1] == "search_query":
                        # extract the search query
                        where["search_query"] = arg.args[1].value
                    else:
                        raise NotImplementedError(
                            f"where clause parameter {arg.args[0].parts[-1]} is not supported, "
                            f"only 'meta_data_filter' and 'search_query' are supported"
                        )
                else:
                    raise NotImplementedError(
                        f"Unsupported where clause {arg.op} operator, only '=' is supported"
                    )

        elif query.where.op == "=":
            # requires separate handling as the where clause is not a list when single condition
            if query.where.args[0].parts[-1] == "meta_data_filter":
                # filters on metadata
                where["meta_data_filter"] = get_metadata_filter(
                    query.where.args[1].value
                )

            elif query.where.args[0].parts[-1] == "search_query":
                # extract the search query
                where["search_query"] = query.where.args[1].value
            else:
                raise NotImplementedError(
                    f"where clause parameter {query.where.args[0].parts[-1]} is not supported, "
                    f"only 'meta_data_filter' and 'search_query' are supported"
                )

        else:
            raise NotImplementedError(
                f"Unsupported where clause {query.where.op} operator, only '=' and 'and' is supported"
            )

        return where

    def query_collection(self, query: ASTNode) -> Response:
        """
        Run a select query on the ChromaDB database, filter collection (if where clause)
        and return result. NB Limit will define the number of documents returned, not the number of rows.
        """

        collection_name = query.from_table.parts[-1]
        self._collection = self._client.get_collection(collection_name)

        if hasattr(query.targets[0], "op") and query.targets[0].op == "count":

            # if count is used, return the count of the collection
            if not isinstance(query.targets[0].args[0], Star):
                # if count is not using '*' argument, raise error
                return Response(
                    resp_type=RESPONSE_TYPE.ERROR,
                    error_message="Count only supports '*' argument",
                )

            collection_data = self._collection.count()
            result = pd.DataFrame(columns=["count"], data=[[collection_data]])

        else:

            # if there is a where clause, parse it to Chromadb query syntax
            where = self.filter_query(query) if query.where else {}

            if where.get("search_query"):
                # if 'search_query' provided, run similarity search

                collection_data = self.similarity_search(
                    filter=where.get(
                        "meta_data_filter"
                    ),  # filters on metadata, for now only column name
                    query=where.get("search_query"),
                    k=query.limit.value if query.limit else 5,
                )
                result = documents_to_df(collection_data)

            else:
                # if no 'search_query' provided, run a regular query and filter on metadata if provided
                collection_data = self._collection.get(
                    where=where.get("meta_data_filter")
                )
                result = pd.DataFrame(collection_data)[
                    ["ids", "documents", "metadatas"]
                ]

        if query.limit:
            # if there is a limit clause, limit the result.
            result = result.head(query.limit.value)

        return Response(resp_type=RESPONSE_TYPE.TABLE, data_frame=result)

    def run_create_table(self, query: ASTNode) -> Response:
        """
        Run a create table query on the ChromaDB database.
        """
        collection_name = query.name.parts[-1]
        self._client.create_collection(collection_name)

        return Response(resp_type=RESPONSE_TYPE.OK)

    def run_insert(self, query):
        """
        Run an insert query on the ChromaDB database.
        """
        collection_name = query.table.parts[-1]
        columns = [column.name for column in query.columns]
        df = pd.DataFrame(data=query.values, columns=columns)
        documents = split_documents(df, columns)

        # converts list of Documents to embedding vectors and stores them in a ChromaDB collection
        Chroma.from_documents(
            documents=documents,
            embedding=self._embedding_function,
            persist_directory=self._persist_directory,
            client_settings=Settings(**self._client_config),
            collection_name=collection_name,
        )

        return Response(resp_type=RESPONSE_TYPE.OK)

    def query(self, query: ASTNode) -> Response:
        """
        Execute a query on the ChromaDB database.
        """
        try:
            self.connect()

            if isinstance(query, Select):
                return self.query_collection(query)

            elif isinstance(query, CreateTable):
                return self.run_create_table(query)

            elif isinstance(query, Insert):
                return self.run_insert(query)

            else:
                raise NotImplementedError(
                    f"Unsupported query type {query.__class__.__name__}!"
                )

        except Exception as e:
            log.logger.error(f"Error executing query on ChromaDB client, {e}!")
            return Response(
                resp_type=RESPONSE_TYPE.ERROR,
                error_message=f"Error executing query on ChromaDB client, {e}!",
            )

    def native_query(self, query: str) -> Response:
        """
        Execute a native query on the ChromaDB database.
        """

        # todo fix this - not sure if we need this?
        try:
            self.connect()
            # parse query and extract collection name and any conditions
            collection_name = extract_collection_name(query)
            self._collection = self._client.get_collection(collection_name)

        except Exception as e:
            log.logger.error(
                f'Error executing native query on ChromaDB client {self._connection_data["chroma_server_host"]}, {e}!'
            )
            return Response(
                resp_type=RESPONSE_TYPE.ERROR,
                error_message=f'Error executing native query on ChromaDB client {self._connection_data["chroma_server_host"]}, {e}!',
            )

    def get_tables(self) -> Response:
        """Get the list of indexes/collections in the vectorDB.

        Returns:
           Response: The response object.
        """

        try:
            self.connect()
            collections = self._client.list_collections()
            collections_name = pd.DataFrame(
                columns=["table_name"],
                data=[collection.name for collection in collections],
            )
        except Exception as e:
            log.logger.error(
                f'Error getting tables from ChromaDB client {self._connection_data["chroma_server_host"]}, {e}!'
            )
            return Response(
                resp_type=RESPONSE_TYPE.ERROR,
                error_message=f'Error getting tables from ChromaDB client {self._connection_data["chroma_server_host"]}, {e}!',
            )

        return Response(
            resp_type=RESPONSE_TYPE.TABLE,
            data_frame=collections_name,
        )


connection_args = OrderedDict(
    chroma_api_impl={
        "type": ARG_TYPE.STR,
        "description": "chromadb api implementation",
    },
    chroma_server_host={
        "type": ARG_TYPE.STR,
        "description": "chromadb server host",
    },
    chroma_server_http_port={
        "type": ARG_TYPE.INT,
        "description": "chromadb server port",
    },
    embedding_function={
        "type": ARG_TYPE.STR,
        "description": "embedding function to use (optional)",
    },
)

connection_args_example = OrderedDict(
    chroma_api_impl="rest",
    chroma_server_host="localhost",
    chroma_server_http_port=8000,
    embedding_function="sentence-transformers/all-mpnet-base-v2",
)



import unittest
from mindsdb.integrations.handlers.ingres_handler.ingres_handler import IngresHandler
from mindsdb.api.mysql.mysql_proxy.libs.constants.response_type import RESPONSE_TYPE


class IngresHandlerTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.kwargs = {
            "connection_data": {
                "user": "admin",
                "password": "password",
                "server": "(local)",
                "database": "test_db"
            }
        }
        cls.handler = IngresHandler('test_ingres_handler', **cls.kwargs)

    def test_0_check_connection(self):
        assert self.handler.check_connection()

    def test_1_get_tables(self):
        tables = self.handler.get_tables()
        assert tables.type is not RESPONSE_TYPE.ERROR

    def test_2_select_query(self):
        query = "SELECT * FROM test_db.home_rentals"
        result = self.handler.native_query(query)
        assert result.type is RESPONSE_TYPE.TABLE

    def test_3_get_columns(self):
        columns = self.handler.get_columns('test')
        assert columns.type is not RESPONSE_TYPE.ERROR

    def test_4_drop_table(self):
        res = self.handler.native_query("DROP TABLE IF EXISTS test_db.test")
        assert res.type is not RESPONSE_TYPE.ERROR

    def test_5_create_table(self):
        res = self.handler.native_query("CREATE TABLE IF NOT EXISTS test_db.test (id INT, name VARCHAR(255))")
        assert res.type is not RESPONSE_TYPE.ERROR

    def test_6_insert(self):
        res = self.handler.native_query("INSERT INTO test VALUES (100,'ONE HUNDRED')")
        assert res.type is not RESPONSE_TYPE.ERROR

    def test_7_disconnect(self):
        assert self.handler.disconnect()


if __name__ == '__main__':
    unittest.main()



import threading
import time

from mindsdb.api.mysql.mysql_proxy.controllers.session_controller import SessionController
from mindsdb.interfaces.storage import db
from mindsdb.utilities import log
from mindsdb.utilities.context import context as ctx
from mindsdb.interfaces.chatbot.chatbot_task import ChatBotTask


class ChatBotThread(threading.Thread):
    """A thread for polling style chatbots to operate."""
    def __init__(self, bot_record):
        threading.Thread.__init__(self)
        self.bot_record = bot_record
        self._to_stop = False

    def run(self):
        # create context and session

        ctx.set_default()
        ctx.company_id = self.bot_record.company_id
        if self.bot_record.user_class is not None:
            ctx.user_class = self.bot_record.user_class

        session = SessionController()

        # TODO check deleted, raise errors
        # TODO checks on delete predictor/ project/ integration
        model_name = self.bot_record.model_name
        project_name = db.Project.query.get(self.bot_record.project_id).name

        database_name = db.Integration.query.get(self.bot_record.database_id).name

        task = ChatBotTask(session,
                           database=database_name,
                           project_name=project_name,
                           model_name=model_name,
                           params=self.bot_record.params)

        while True:
            try:
                task.run()
            except Exception as e:
                log.logger.error(e)

            if self._to_stop:
                return
            log.logger.debug('running ' + self.name)
            time.sleep(7)

    def stop(self):
        self._to_stop = True



from mindsdb.api.mongo.classes import Responder
import mindsdb.api.mongo.functions as helpers


class Responce(Responder):
    when = {'whatsmyuri': helpers.is_true}

    def result(self, query, request_env, mindsdb_env, session):
        mongodb_config = mindsdb_env['config']['api']['mongodb']
        host = mongodb_config['host']
        port = mongodb_config['port']
        return {
            'you': f'{host}:{port}',
            'ok': 1
        }


responder = Responce()



import requests
import pandas as pd
from mindsdb.utilities import log
from mindsdb.utilities.config import Config
from mindsdb_sql.parser import ast
from mindsdb.integrations.libs.api_handler import APIHandler, APITable
from mindsdb.integrations.libs.response import HandlerStatusResponse as StatusResponse, HandlerResponse as Response, RESPONSE_TYPE
from .hn_table import StoriesTable, CommentsTable

class HackerNewsHandler(APIHandler):
    """
    A class for handling connections and interactions with the Hacker News API.
    """

    def __init__(self, name=None, **kwargs):
        super().__init__(name)

        self.base_url = 'https://hacker-news.firebaseio.com/v0'

        stories = StoriesTable(self)
        self._register_table('stories', stories)

        comments = CommentsTable(self)
        self._register_table('comments', comments)

    def connect(self):
        return

    def check_connection(self) -> StatusResponse:
        try:
            response = requests.get(f'{self.base_url}/maxitem.json')
            response.raise_for_status()
            return StatusResponse(True)
        except Exception as e:
            log.error(f'Error checking connection: {e}')
            return StatusResponse(False, str(e))

    def native_query(self, query_string: str = None):
        method_name, params = self.parse_native_query(query_string)

        df = self.call_hackernews_api(method_name, params)

        return Response(
            RESPONSE_TYPE.TABLE,
            data_frame=df
        )

    def call_hackernews_api(self, method_name: str = None, params: dict = None):
            if method_name == 'get_top_stories':
                url = f'{self.base_url}/topstories.json'
                response = requests.get(url)
                data = response.json()
                stories_data = []
                for story_id in data:
                    url = f'{self.base_url}/item/{story_id}.json'
                    response = requests.get(url)
                    story_data = response.json()
                    stories_data.append(story_data)
                df = pd.DataFrame(stories_data, columns=['id', 'time', 'title', 'url', 'score', 'descendants'])
            elif method_name == 'get_comments':
                item_id = params.get('item_id')
                url = f'{self.base_url}/item/{item_id}.json'
                response = requests.get(url)
                item_data = response.json()
                if 'kids' in item_data:
                    comments_data = []
                    for comment_id in item_data['kids']:
                        url = f'{self.base_url}/item/{comment_id}.json'
                        response = requests.get(url)
                        comment_data = response.json()
                        comments_data.append(comment_data)
                    df = pd.DataFrame(comments_data)
                else:
                    df = pd.DataFrame()
            else:
                raise ValueError(f'Unknown method_name: {method_name}')

            return df




import unittest
from mindsdb.integrations.handlers.cockroach_handler.cockroach_handler import CockroachHandler
from mindsdb.api.mysql.mysql_proxy.libs.constants.response_type import RESPONSE_TYPE


class CockroachHandlerTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.kwargs = {
            "connection_data": {
                "host": "localhost",
                "port": "5432",
                "user": "mindsdb",
                "password": "mindsdb",
                "database": "postgres"
            }
        }
        cls.handler = CockroachHandler('test_cockroach_handler', **cls.kwargs)

    def test_0_check_connection(self):
        assert self.handler.check_connection()

    def test_1_describe_table(self):
        described = self.handler.describe_table("test_mdb")
        assert described['type'] is not RESPONSE_TYPE.ERROR

    def test_2_get_tables(self):
        tables = self.handler.get_tables()
        assert tables['type'] is not RESPONSE_TYPE.ERROR

    def test_4_select_query(self):
        query = "SELECT * FROM data.test_mdb WHERE 'id'='1'"
        result = self.handler.query(query)
        assert result['type'] is RESPONSE_TYPE.TABLE



"""
*******************************************************
 * Copyright (C) 2017 MindsDB Inc. <copyright@mindsdb.com>
 *
 * This file is part of MindsDB Server.
 *
 * MindsDB Server can not be copied and/or distributed without the express
 * permission of MindsDB Inc
 *******************************************************
"""

import os
from uuid import uuid4

import requests

from mindsdb.api.mysql.mysql_proxy.datahub import init_datahub
from mindsdb.api.mysql.mysql_proxy.utilities import logger
from mindsdb.utilities.config import Config
from mindsdb.interfaces.model.model_controller import ModelController
from mindsdb.interfaces.database.database import DatabaseController
from mindsdb.interfaces.database.integrations import integration_controller


class SessionController:
    """
    This class manages the server session
    """

    def __init__(self) -> object:
        """
        Initialize the session
        """

        self.username = None
        self.auth = False
        self.logging = logger
        self.database = None

        self.config = Config()

        self.model_controller = ModelController()
        self.integration_controller = integration_controller
        self.database_controller = DatabaseController()

        self.datahub = init_datahub(self)

        self.prepared_stmts = {}
        self.packet_sequence_number = 0
        self.profiling = False
        self.predictor_cache = True

    def inc_packet_sequence_number(self):
        self.packet_sequence_number = (self.packet_sequence_number + 1) % 256

    def register_stmt(self, statement):
        i = 1
        while i in self.prepared_stmts and i < 100:
            i = i + 1
        if i == 100:
            raise Exception("Too many unclosed queries")

        self.prepared_stmts[i] = dict(type=None, statement=statement, fetched=0)
        return i

    def unregister_stmt(self, stmt_id):
        del self.prepared_stmts[stmt_id]

    def to_json(self):
        return {
            "username": self.username,
            "auth": self.auth,
            "database": self.database,
            "prepared_stmts": self.prepared_stmts,
            "packet_sequence_number": self.packet_sequence_number,
        }

    def from_json(self, updated):
        for key in updated:
            setattr(self, key, updated[key])


class ServerSessionContorller(SessionController):
    """SessionController implementation for case of Executor service.
    The difference with SessionController is that there is an id in this one.
    The instance uses the id to synchronize its settings with the appropriate
    ServiceSessionController instance on the Executor side."""

    def __init__(self):
        super().__init__()
        self.id = f"session_{uuid4()}"
        self.executor_url = os.environ.get("MINDSDB_EXECUTOR_URL", None)
        self.executor_host = os.environ.get("MINDSDB_EXECUTOR_SERVICE_HOST", None)
        self.executor_port = os.environ.get("MINDSDB_EXECUTOR_SERVICE_PORT", None)
        if (self.executor_host is None or self.executor_port is None) and self.executor_url is None:
            raise Exception(f"""{self.__class__.__name__} can be used only in modular mode of MindsDB.
                            Use Executor as a service and specify MINDSDB_EXECUTOR_URL env variable""")
        logger.info(
            "%s.__init__: executor url - %s", self.__class__.__name__, self.executor_url
        )

    def __del__(self):
        """Terminate the appropriate ServiceSessionController instance as well."""
        if self.executor_url is not None:
            url = self.executor_url + "/" + "session"
            logger.info(
                "%s.__del__: delete an appropriate ServiceSessionController with, id - %s on the Executor service side",
                self.__class__.__name__,
                self.id,
            )
            requests.delete(url, json={"id": self.id})



"""
*******************************************************
 * Copyright (C) 2017 MindsDB Inc. <copyright@mindsdb.com>
 *
 * This file is part of MindsDB Server.
 *
 * MindsDB Server can not be copied and/or distributed without the express
 * permission of MindsDB Inc
 *******************************************************
"""
import datetime as dt
import struct

import pandas as pd

from mindsdb.api.mysql.mysql_proxy.data_types.mysql_datum import Datum
from mindsdb.api.mysql.mysql_proxy.data_types.mysql_packet import Packet
from mindsdb.api.mysql.mysql_proxy.libs.constants.mysql import TYPES


class BinaryResultsetRowPacket(Packet):
    '''
    Implementation based on:
    https://mariadb.com/kb/en/resultset-row/#binary-resultset-row
    https://dev.mysql.com/doc/internals/en/null-bitmap.html
    '''

    def setup(self):
        data = self._kwargs.get('data', {})
        columns = self._kwargs.get('columns', {})

        self.value = [b'\x00']
        nulls = [0]
        for i, el in enumerate(data):
            if i > 0 and (i + 2) % 8 == 0:
                nulls.append(0)
            if el is None:
                if i < 6:
                    nulls[-1] = nulls[-1] + (1 << ((i + 2) % 8))
                else:
                    nulls[-1] = nulls[-1] + (1 << ((i - 6) % 8))
        self.value.append(bytes(nulls))

        for i, col in enumerate(columns):
            # NOTE at this moment all types sends as strings, and it works
            val = data[i]
            if val is None:
                continue

            enc = None
            env_val = None
            col_type = col['type']
            if col_type == TYPES.MYSQL_TYPE_DOUBLE:
                enc = '<d'
                val = float(val)
            elif col_type == TYPES.MYSQL_TYPE_LONGLONG:
                enc = '<q'
                val = int(float(val))
            elif col_type == TYPES.MYSQL_TYPE_LONG:
                enc = '<l'
                val = int(float(val))
            elif col_type == TYPES.MYSQL_TYPE_FLOAT:
                enc = '<f'
                val = float(val)
            elif col_type == TYPES.MYSQL_TYPE_YEAR:
                enc = '<h'
                val = int(float(val))
            elif col_type == TYPES.MYSQL_TYPE_DATE:
                env_val = self.encode_date(val)
            elif col_type == TYPES.MYSQL_TYPE_TIMESTAMP:
                env_val = self.encode_date(val)
            elif col_type == TYPES.MYSQL_TYPE_DATETIME:
                env_val = self.encode_date(val)
            elif col_type == TYPES.MYSQL_TYPE_TIME:
                enc = ''
            elif col_type == TYPES.MYSQL_TYPE_NEWDECIMAL:
                enc = ''
            else:
                enc = 'string'

            if enc == '':
                raise Exception(f'Column with type {col_type} cant be encripted')

            if enc == 'string':
                self.value.append(Datum('string<lenenc>', str(val)).toStringPacket())
            else:
                if env_val is None:
                    env_val = struct.pack(enc, val)
                self.value.append(env_val)

    def encode_date(self, val):
        # date_type = None
        # date_value = None

        if isinstance(val, str):
            try:
                date_value = dt.datetime.strptime(val, '%Y-%m-%d')
                date_type = 'date'
            except ValueError:
                try:
                    date_value = dt.datetime.strptime(val, '%Y-%m-%dT%H:%M:%S')
                    date_type = 'datetime'
                except ValueError:
                    date_value = dt.datetime.strptime(val, '%Y-%m-%dT%H:%M:%S.%f')
                    date_type = 'datetime'
        elif isinstance(val, pd.Timestamp):
            date_value = val
            date_type = 'datetime'

        out = struct.pack('<H', date_value.year)
        out += struct.pack('<B', date_value.month)
        out += struct.pack('<B', date_value.day)

        if date_type == 'datetime':
            out += struct.pack('<B', date_value.hour)
            out += struct.pack('<B', date_value.minute)
            out += struct.pack('<B', date_value.second)
            out += struct.pack('<L', date_value.microsecond)

        len_bit = struct.pack('<B', len(out))
        return len_bit + out

    @property
    def body(self):
        string = b''.join(self.value)
        self.setBody(string)
        return self._body

    @staticmethod
    def test():
        import pprint
        pprint.pprint(
            str(BinaryResultsetRowPacket().get_packet_string())
        )


if __name__ == "__main__":
    BinaryResultsetRowPacket.test()



import random
import string
from typing import Optional
import os
import pandas as pd
import numpy as np
import autokeras as ak
from sklearn import preprocessing
from mindsdb.integrations.libs.base import BaseMLEngine
from tensorflow.keras.models import load_model


# Makes this run on systems where this arg isn't specified, like Windows Subsystem for Linux
# Doesn't break things on Mac
os.environ["XLA_FLAGS"] = "--xla_gpu_cuda_data_dir=/usr/lib/cuda"

trainer_dict = {"regression": ak.StructuredDataRegressor, "classification": ak.StructuredDataClassifier}

DEFAULT_TRIALS = 100


def train_model(df, target, max_trials=DEFAULT_TRIALS):
    """Helper function to trains an AutoKeras model with an input df.

    Automatically decides on classification vs. regression depending on
    the type of the target.

    Will auto-encode categorical variables as dummies.

    Returns both trained model and the names of categoric dummy columns, which
    are passed later to the prediction method.
    """
    # Choose regressor of classifier based on target data type
    if np.issubdtype(df[target].dtype, np.number):
        mode = "regression"
        y_train = df[target]
    else:
        mode = "classification"
        lb = preprocessing.LabelBinarizer()
        y_train = lb.fit_transform(df[target])

    training_df = df.drop(target, axis=1)
    trainer = trainer_dict[mode](overwrite=True, max_trials=max_trials)

    # Save the column names of all numeric columns before transforming any categorical columns into dummies
    numeric_column_names = training_df.select_dtypes(include=[np.number]).columns.values.tolist()
    training_df = pd.get_dummies(training_df)
    categorical_dummy_column_names = [
        col for col in training_df.columns.values.tolist() if col not in numeric_column_names
    ]
    trainer.fit(training_df, y_train, verbose=2)
    return trainer.export_model(), categorical_dummy_column_names


def get_preds_from_model(df, model, target, column_count, categorical_dummy_column_names):
    """Gets predictions from the stored AutoKeras model."""
    df_to_predict = df.copy()
    for col in ["__mindsdb_row_id", target]:
        if col in df_to_predict.columns.values.tolist():
            df_to_predict = df_to_predict.drop(col, axis=1)

    if len(df_to_predict.columns) != column_count:
        raise Exception("All feature columns must be defined in the WHERE clause when making predictions")
    # Get dummies for any categorical columns and then populate the missing ones with zeros
    df_with_dummies = pd.get_dummies(df_to_predict)
    for col in categorical_dummy_column_names:  # exception handler for empty columns
        if col not in df_with_dummies.columns.values.tolist():
            df_with_dummies[col] = 0

    return model.predict(df_with_dummies, verbose=2)


def format_categorical_preds(predictions, original_y, df_to_predict, target_col):
    """Transforms class predictions back to their original class.

    Categoric predictions come out the AutoKeras model in a binary
    format e.g. (0, 1, 0). This function maps them back to their
    original class e.g. 'Blue', and adds a DF column for the
    model confidence score.
    """
    # Turn prediction back into categorical value
    lb = preprocessing.LabelBinarizer()
    lb.fit(original_y)
    preds = lb.inverse_transform(predictions)

    # Add the confidence score next to the prediction
    df_to_predict[target_col] = pd.Series(preds).astype(original_y.dtype)
    df_to_predict["confidence"] = [max(row) for _, row in enumerate(predictions)]
    return df_to_predict


class AutokerasHandler(BaseMLEngine):
    """
    Integration with the AutoKeras ML library.
    """  # noqa

    name = "autokeras"

    def create(self, target: str, df: Optional[pd.DataFrame] = None, args: Optional[dict] = None) -> None:
        """Create and tune AutoKeras model using the input df.

        Saves the AutoKeras best model params to model storage.
        """
        args = args["using"]  # ignore the rest of the problem definition
        args["target"] = target
        max_trials = int(args["train_time"] * DEFAULT_TRIALS) if "train_time" in args else DEFAULT_TRIALS
        # Save the training df in order to filter the training data based on the predict df
        args["training_df"] = df.to_json()
        args["training_data_column_count"] = len(df.columns) - 1  # subtract 1 for target

        random_string = "".join(random.choices(string.ascii_uppercase + string.digits, k=24))
        args["folder_path"] = os.path.join("autokeras", random_string)

        model, args["data_column_names"] = train_model(df, target, max_trials)
        model.save(args["folder_path"])
        self.model_storage.json_set("predict_args", args)

    def predict(self, df, args=None):
        """Predicts with best saved AutoKeras model."""
        args = self.model_storage.json_get("predict_args")
        training_df = pd.read_json(args["training_df"])
        model = load_model(args["folder_path"], custom_objects=ak.CUSTOM_OBJECTS)

        df_to_predict = df.copy()
        predictions = get_preds_from_model(df_to_predict, model, args["target"], args["training_data_column_count"], args["data_column_names"])

        # If we used the classifier we need to pre-process the predictions before returning them
        original_y = training_df[args["target"]]
        if not np.issubdtype(original_y.dtype, np.number):
            return format_categorical_preds(predictions, original_y, df_to_predict, args["target"])

        df_to_predict[args["target"]] = predictions
        return df_to_predict



"""
How to use it:

    from mindsdb.utilities.cache import get_cache, dataframe_checksum, json_checksum

    # namespace of cache
    cache = get_cache('predict')

    key = dataframe_checksum(df) # or json_checksum, depends on object type
    df_predict = cache(key)

    if df_predict is None:
        # no cache, save it
        df_predict = predictor.predict(df)
        cache.set(key, df_predict)



Configuration:

- max_size size of cache in count of records, default is 50
- serializer, module for serialization, default is dill

It can be set via:
- get_cache function:
    cache = get_cache('predict', max_size=2)
- using specific cache class:
    cache = FileCache('predict', max_size=2)
- using mindsdb config file:
    "cache": {
        "type": "redis",
        "max_size": 2
    }

Cache engines:

Can be specified in mindsdb config json. Possible values:
- local - for FileCache, default
- redis - for RedisCache
By default is used local redis server. You can specify
    "cache": {
        "type": "redis",
        "connection": {
            "host": "127.0.0.1",
            "port": 6379
        }
    }

How to test:

    env PYTHONPATH=./ pytest tests/unit/test_cache.py

"""

import os
import time
from abc import ABC
from pathlib import Path
import hashlib
import typing as t

import pandas as pd
import walrus

from mindsdb.utilities.config import Config
from mindsdb.utilities.json_encoder import CustomJSONEncoder
from mindsdb.interfaces.storage.fs import FileLock
from mindsdb.utilities.context import context as ctx


def dataframe_checksum(df: pd.DataFrame):
    checksum = str_checksum(df.to_json())
    return checksum


def json_checksum(obj: t.Union[dict, list]):
    checksum = str_checksum(CustomJSONEncoder().encode(obj))
    return checksum


def str_checksum(obj: str):
    checksum = hashlib.sha256(obj.encode()).hexdigest()
    return checksum


class BaseCache(ABC):
    def __init__(self, max_size=None, serializer=None):
        self.config = Config()
        if max_size is None:
            max_size = self.config["cache"].get("max_size", 50)
        self.max_size = max_size
        if serializer is None:
            serializer_module = self.config["cache"].get('serializer')
            if serializer_module == 'pickle':
                import pickle as s_module
            else:
                import dill as s_module
            self.serializer = s_module

    # default functions

    def set_df(self, name, df):
        return self.set(name, df)

    def get_df(self, name):
        return self.get(name)

    def serialize(self, value):
        return self.serializer.dumps(value)

    def deserialize(self, value):
        return self.serializer.loads(value)


class FileCache(BaseCache):
    def __init__(self, category, path=None, **kwargs):
        super().__init__(**kwargs)

        if path is None:
            path = self.config['paths']['cache']

        cache_path = Path(path) / category

        company_id = ctx.company_id
        if company_id is not None:
            cache_path = cache_path / str(company_id)
        cache_path.mkdir(parents=True, exist_ok=True)

        self.path = cache_path

    def clear_old_cache(self):
        with FileLock(self.path):
            # buffer to delete, to not run delete on every adding
            buffer_size = 5

            if self.max_size is None:
                return

            cur_count = len(os.listdir(self.path))

            if cur_count > self.max_size + buffer_size:
                files = sorted(Path(self.path).iterdir(), key=os.path.getmtime)
                for file in files[:cur_count - self.max_size]:
                    self.delete_file(file)

    def file_path(self, name):
        return self.path / name

    def set_df(self, name, df):
        path = self.file_path(name)
        df.to_pickle(path)
        self.clear_old_cache()

    def set(self, name, value):
        path = self.file_path(name)
        value = self.serialize(value)

        with open(path, 'wb') as fd:
            fd.write(value)
        self.clear_old_cache()

    def get_df(self, name):
        path = self.file_path(name)
        with FileLock(self.path):
            if not os.path.exists(path):
                return None
            value = pd.read_pickle(path)
        return value

    def get(self, name):
        path = self.file_path(name)

        with FileLock(self.path):
            if not os.path.exists(path):
                return None
            with open(path, 'rb') as fd:
                value = fd.read()
        value = self.deserialize(value)
        return value

    def delete(self, name):
        path = self.file_path(name)
        self.delete_file(path)

    def delete_file(self, path):
        os.unlink(path)


class RedisCache(BaseCache):
    def __init__(self, category, connection_info=None, **kwargs):
        super().__init__(**kwargs)

        self.category = category

        if connection_info is None:
            # if no params will be used local redis
            connection_info = self.config["cache"].get("connection", {})
        self.client = walrus.Database(**connection_info)

    def clear_old_cache(self, key_added):

        if self.max_size is None:
            return

        # buffer to delete, to not run delete on every adding
        buffer_size = 5

        cur_count = self.client.hlen(self.category)

        # remove oldest
        if cur_count > self.max_size + buffer_size:
            # 5 is buffer to delete, to not run delete on every adding

            keys = self.client.hgetall(self.category)
            # to list
            keys = list(keys.items())
            # sort by timestamp
            keys.sort(key=lambda x: x[1])

            for key, _ in keys[:cur_count - self.max_size]:
                self.delete_key(key)

    def redis_key(self, name):
        return f'{self.category}_{name}'

    def set(self, name, value):
        key = self.redis_key(name)
        value = self.serialize(value)

        self.client.set(key, value)
        # using key with category name to store all keys with modify time
        self.client.hset(self.category, key, int(time.time() * 1000))

        self.clear_old_cache(key)

    def get(self, name):
        key = self.redis_key(name)
        value = self.client.get(key)
        if value is None:
            # no value in cache
            return None
        return self.deserialize(value)

    def delete(self, name):
        key = self.redis_key(name)

        self.delete_key(key)

    def delete_key(self, key):
        self.client.delete(key)
        self.client.hdel(self.category, key)


class NoCache:
    '''
        class for no cache mode
    '''
    def __init__(self, *args, **kwargs):
        pass

    def get(self, name):
        return None

    def set(self, name, value):
        pass


def get_cache(category, **kwargs):
    config = Config()
    if config.get('cache')['type'] == 'redis':
        return RedisCache(category, **kwargs)
    if config.get('cache')['type'] == 'none':
        return NoCache(category, **kwargs)
    else:
        return FileCache(category, **kwargs)



from typing import List

from mindsdb.integrations.libs.api_handler import APITable
from mindsdb.integrations.utilities.sql_utils import extract_comparison_conditions
from mindsdb.utilities.log import get_log

from mindsdb_sql.parser import ast


from googleapiclient.discovery import build
from googleapiclient.errors import HttpError
import os
import pandas as pd


logger = get_log("integrations.youtube_handler")

class YoutubeGetCommentsTable(APITable):
    """Youtube List Comments  by video id Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Pulls data from the youtube  "commentThreads()" API endpoint
        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query
        Returns
        -------
        pd.DataFrame
            youtube "commentThreads()" matching the query
        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """
        conditions = extract_comparison_conditions(query.where)

        order_by_conditions = {}
        clubs_kwargs = {}

        if query.order_by and len(query.order_by) > 0:
            order_by_conditions["columns"] = []
            order_by_conditions["ascending"] = []

            for an_order in query.order_by:
                if an_order.field.parts[0] != "id":
                    next    
                if an_order.field.parts[1] in self.get_columns():
                    order_by_conditions["columns"].append(an_order.field.parts[1])

                    if an_order.direction == "ASC":
                        order_by_conditions["ascending"].append(True)
                    else:
                        order_by_conditions["ascending"].append(False)
                else:
                    raise ValueError(
                        f"Order by unknown column {an_order.field.parts[1]}"
                    )

        for a_where in conditions:
            if a_where[1] == "youtube_video_id":
                if a_where[0] != "=":
                    raise ValueError("Unsupported where operation for youtube video id")
                clubs_kwargs["type"] = a_where[2]
            else:
                raise ValueError(f"Unsupported where argument {a_where[1]}")


        youtube_comments_df = self.call_youtube_comments_api(a_where[2])

        selected_columns = []
        for target in query.targets:
            if isinstance(target, ast.Star):
                selected_columns = self.get_columns()
                break
            elif isinstance(target, ast.Identifier):
                selected_columns.append(target.parts[-1])
            else:
                raise ValueError(f"Unknown query target {type(target)}")


        if len(youtube_comments_df) == 0:
            youtube_comments_df = pd.DataFrame([], columns=selected_columns)
        else:
            youtube_comments_df.columns = self.get_columns()
            for col in set(youtube_comments_df.columns).difference(set(selected_columns)):
                youtube_comments_df = youtube_comments_df.drop(col, axis=1)

            if len(order_by_conditions.get("columns", [])) > 0:
                youtube_comments_df = youtube_comments_df.sort_values(
                    by=order_by_conditions["columns"],
                    ascending=order_by_conditions["ascending"],
                )

        if query.limit:
            youtube_comments_df = youtube_comments_df.head(query.limit.value)

        return youtube_comments_df

    def get_columns(self) -> List[str]:
        """Gets all columns to be returned in pandas DataFrame responses
        Returns
        -------
        List[str]
            List of columns
        """
        return [
        'user_id', 
        'display_name', 
        'comment'
        ]

    def call_youtube_comments_api(self,video_id):
        """Pulls all the records from the given youtube api end point and returns it select()
    
        Returns
        -------
        pd.DataFrame of all the records of the "commentThreads()" API end point
        """

        resource = self.handler.connect().commentThreads().list(part='snippet',videoId=video_id,textFormat='plainText')

        video_cols = self.get_columns()
        all_youtube_comments_df = pd.DataFrame(columns=video_cols)
        data = []

        while resource:
            comments = resource.execute()
            for comment in comments['items']:
                user_id = comment['snippet']['topLevelComment']['snippet']['authorChannelId']['value']
                display_name = comment['snippet']['topLevelComment']['snippet']['authorDisplayName']
                comment_text = comment['snippet']['topLevelComment']['snippet']['textDisplay']
                data = pd.DataFrame({
                    'user_id': [user_id],
                    'display_name': [display_name],
                    'comment': [comment_text]
                })
                all_youtube_comments_df = pd.concat([all_youtube_comments_df, data], ignore_index=True)
            if 'nextPageToken' in comments:
                resource = self.handler.connect().commentThreads().list(
                    part='snippet',
                    videoId=video_id,
                    textFormat='plainText',
                    pageToken=comments['nextPageToken']
                )
            else:
                break


        return all_youtube_comments_df



from http import HTTPStatus
from sqlalchemy.exc import NoResultFound
from typing import Dict

from flask import request
from flask_restx import Resource, abort

from mindsdb.api.http.namespaces.configs.databases import ns_conf
from mindsdb.api.mysql.mysql_proxy.classes.fake_mysql_proxy import FakeMysqlProxy
from mindsdb.api.mysql.mysql_proxy.controllers.session_controller import SessionController
from mindsdb.api.mysql.mysql_proxy.datahub.classes.tables_row import TablesRow
from mindsdb.api.mysql.mysql_proxy.libs.constants.response_type import RESPONSE_TYPE
from mindsdb_sql import parse_sql, ParsingException
from mindsdb_sql.parser.ast import CreateTable, DropTables


@ns_conf.route('/')
class DatabasesResource(Resource):
    @ns_conf.doc('list_databases')
    def get(self):
        '''List all databases'''
        session = SessionController()
        return session.database_controller.get_list()

    @ns_conf.doc('create_database')
    def post(self):
        '''Create a database'''
        if 'database' not in request.json:
            abort(HTTPStatus.BAD_REQUEST, 'Must provide "database" parameter in POST body')
        session = SessionController()
        database = request.json['database']
        parameters = {}
        if 'name' not in database:
            abort(HTTPStatus.BAD_REQUEST, 'Missing "name" field for database')
        if 'engine' not in database:
            abort(HTTPStatus.BAD_REQUEST, 'Missing "engine" field for database. If you want to create a project instead, use the /api/projects endpoint.')
        if 'parameters' in database:
            parameters = database['parameters']
        name = database['name']

        if session.database_controller.exists(name):
            abort(HTTPStatus.CONFLICT, f'Database with name {name} already exists.')
        new_integration_id = session.integration_controller.add(name, database['engine'], parameters)
        new_integration = session.database_controller.get_integration(new_integration_id)
        return new_integration, HTTPStatus.CREATED


@ns_conf.route('/<database_name>')
class DatabaseResource(Resource):
    @ns_conf.doc('get_database')
    def get(self, database_name):
        '''Gets a database by name'''
        session = SessionController()
        try:
            project = session.database_controller.get_project(database_name)
            return {
                'name': database_name,
                'type': 'project',
                'id': project.id,
                'engine': None
            }
        except NoResultFound:
            integration = session.integration_controller.get(database_name)
            if integration is None:
                abort(HTTPStatus.NOT_FOUND, f'Database with name {database_name} does not exist.')
            return integration

    @ns_conf.doc('update_database')
    def put(self, database_name):
        '''Updates or creates a database'''
        if 'database' not in request.json:
            abort(HTTPStatus.BAD_REQUEST, 'Must provide "database" parameter in POST body')

        session = SessionController()
        parameters = {}
        database = request.json['database']
        if 'parameters' in database:
            parameters = database['parameters']
        if not session.database_controller.exists(database_name):
            # Create.
            if 'engine' not in database:
                abort(HTTPStatus.BAD_REQUEST, 'Missing "engine" field for new database. If you want to create a project instead, use the POST /api/projects endpoint.')
            new_integration_id = session.integration_controller.add(database_name, database['engine'], parameters)
            new_integration = session.database_controller.get_integration(new_integration_id)
            return new_integration, HTTPStatus.CREATED

        session.integration_controller.modify(database_name, parameters)
        return session.integration_controller.get(database_name)

    @ns_conf.doc('delete_database')
    def delete(self, database_name):
        '''Deletes a database by name'''
        session = SessionController()
        if not session.database_controller.exists(database_name):
            abort(HTTPStatus.NOT_FOUND, f'Database with name {database_name} does not exist.')
        try:
            session.database_controller.delete(database_name)
        except Exception as e:
            abort(
                HTTPStatus.BAD_REQUEST,
                f'Cannot delete database {database_name}. '
                + 'This is most likely a system database, a permanent integration, or an ML engine with active models. '
                + f'Full error: {e}. '
                + 'Please check the name and try again.')
        return '', HTTPStatus.NO_CONTENT


def _tables_row_to_obj(table_row: TablesRow) -> Dict:
    type = table_row.TABLE_TYPE.lower()
    if table_row.TABLE_TYPE == 'BASE TABLE':
        type = 'data'
    return {
        'name': table_row.TABLE_NAME,
        'type': type
    }


@ns_conf.route('/<database_name>/tables')
class TablesList(Resource):
    @ns_conf.doc('list_tables')
    def get(self, database_name):
        '''Get all tables in a database'''
        session = SessionController()
        datanode = session.datahub.get(database_name)
        all_tables = datanode.get_tables()
        table_objs = [_tables_row_to_obj(t) for t in all_tables]
        return table_objs

    @ns_conf.doc('create_table')
    def post(self, database_name):
        '''Creates a table in a database'''
        if 'table' not in request.json:
            abort(HTTPStatus.BAD_REQUEST, 'Must provide "table" parameter in POST body')
        table = request.json['table']
        if 'name' not in table:
            abort(HTTPStatus.BAD_REQUEST, 'Missing "name" field for table')
        if 'select' not in table:
            abort(HTTPStatus.BAD_REQUEST, 'Missing "select" field for table')
        table_name = table['name']
        select_query = table['select']
        replace = False
        if 'replace' in table:
            replace = table['replace']

        session = SessionController()
        try:
            session.database_controller.get_project(database_name)
            error_message = f'Database {database_name} is a project. ' \
                + f'If you want to create a model or view, use the projects/{database_name}/models/{table_name} or ' \
                + f'projects/{database_name}/views/{table_name} endpoints instead.'
            abort(HTTPStatus.BAD_REQUEST, error_message)
        except NoResultFound:
            # Only support creating tables from integrations.
            pass

        datanode = session.datahub.get(database_name)
        if datanode is None:
            abort(HTTPStatus.NOT_FOUND, f'Database with name {database_name} does not exist')
        all_tables = datanode.get_tables()
        for t in all_tables:
            if t.TABLE_NAME == table_name and not replace:
                abort(HTTPStatus.CONFLICT, f'Table with name {table_name} already exists')

        try:
            select_ast = parse_sql(select_query, dialect='mindsdb')
        except ParsingException:
            abort(HTTPStatus.BAD_REQUEST, f'Could not parse select statement {select_query}')

        create_ast = CreateTable(
            f'{database_name}.{table_name}',
            from_select=select_ast,
            is_replace=replace
        )

        mysql_proxy = FakeMysqlProxy()

        try:
            mysql_proxy.process_query(create_ast.get_string())
        except Exception as e:
            abort(HTTPStatus.BAD_REQUEST, e)

        all_tables = datanode.get_tables()
        try:
            matching_table = next(t for t in all_tables if t.TABLE_NAME == table_name)
            return _tables_row_to_obj(matching_table), HTTPStatus.CREATED
        except StopIteration:
            abort(HTTPStatus.INTERNAL_SERVER_ERROR, f'Table with name {table_name} could not be created')


@ns_conf.route('/<database_name>/tables/<table_name>')
@ns_conf.param('database_name', 'Name of the database')
@ns_conf.param('table_name', 'Name of the table')
class TableResource(Resource):
    @ns_conf.doc('get_table')
    def get(self, database_name, table_name):
        session = SessionController()
        datanode = session.datahub.get(database_name)
        all_tables = datanode.get_tables()
        try:
            matching_table = next(t for t in all_tables if t.TABLE_NAME == table_name)
            return _tables_row_to_obj(matching_table)
        except StopIteration:
            abort(HTTPStatus.NOT_FOUND, f'Table with name {table_name} not found')

    @ns_conf.doc('drop_table')
    def delete(self, database_name, table_name):
        session = SessionController()
        try:
            session.database_controller.get_project(database_name)
            error_message = f'Database {database_name} is a project. ' \
                + f'If you want to delete a model or view, use the projects/{database_name}/models/{table_name} or ' \
                + f'projects/{database_name}/views/{table_name} endpoints instead.'
            abort(HTTPStatus.BAD_REQUEST, error_message)
        except NoResultFound:
            # Only support dropping tables from integrations.
            pass

        datanode = session.datahub.get(database_name)
        if datanode is None:
            abort(HTTPStatus.NOT_FOUND, f'Database with name {database_name} not found')
        all_tables = datanode.get_tables()
        try:
            next(t for t in all_tables if t.TABLE_NAME == table_name)
        except StopIteration:
            abort(HTTPStatus.NOT_FOUND, f'Table with name {table_name} not found')

        drop_ast = DropTables(
            tables=[table_name],
            if_exists=True
        )

        try:
            integration_handler = session.integration_controller.get_handler(database_name)
        except Exception:
            abort(HTTPStatus.INTERNAL_SERVER_ERROR, f'Could not get database handler for {database_name}')
        try:
            result = integration_handler.query(drop_ast)
        except NotImplementedError:
            abort(HTTPStatus.BAD_REQUEST, f'Database {database_name} does not support dropping tables.')
        if result.type == RESPONSE_TYPE.ERROR:
            abort(HTTPStatus.BAD_REQUEST, result.error_message)
        return '', HTTPStatus.NO_CONTENT



import datetime
from typing import Optional
from pathlib import Path
from functools import reduce

import pandas as pd
from mindsdb_sql.parser.dialects.mindsdb import (
    CreateDatabase,
    RetrainPredictor,
    CreatePredictor,
    FinetunePredictor,
    CreateMLEngine,
    DropMLEngine,
    DropDatasource,
    DropPredictor,
    CreateView,
    CreateJob,
    DropJob,
    Evaluate,
    CreateChatBot,
    DropChatBot,
)
from mindsdb_sql import parse_sql
from mindsdb_sql.parser.dialects.mysql import Variable
from mindsdb_sql.parser.ast import (
    RollbackTransaction,
    CommitTransaction,
    StartTransaction,
    BinaryOperation,
    DropDatabase,
    NullConstant,
    NativeQuery,
    Describe,
    Constant,
    Function,
    Explain,
    Delete,
    Insert,
    Select,
    Star,
    Show,
    Set,
    Use,
    Alter,
    Update,
    CreateTable,
    Identifier,
    DropTables,
    Operation,
    ASTNode,
    DropView,
    Union,
)
from mindsdb_sql.render.sqlalchemy_render import SqlalchemyRender

from mindsdb_evaluator.accuracy.general import evaluate_accuracy

from mindsdb.api.mysql.mysql_proxy.utilities.sql import query_df
from mindsdb.api.mysql.mysql_proxy.utilities import log
from mindsdb.api.mysql.mysql_proxy.utilities import (
    SqlApiException,
    ErBadDbError,
    ErBadTableError,
    ErTableExistError,
    ErNotSupportedYet,
    ErSqlWrongArguments,
)
from mindsdb.api.mysql.mysql_proxy.utilities.functions import download_file
from mindsdb.api.mysql.mysql_proxy.classes.sql_query import SQLQuery, Column
from mindsdb.api.mysql.mysql_proxy.libs.constants.mysql import (
    CHARSET_NUMBERS,
    TYPES,
    SERVER_VARIABLES,
)
from mindsdb.api.mysql.mysql_proxy.executor.data_types import ExecuteAnswer, ANSWER_TYPE
from mindsdb.integrations.libs.response import HandlerStatusResponse
from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE
from mindsdb.interfaces.model.functions import (
    get_model_record,
    get_model_records,
    get_predictor_integration,
)
from mindsdb.integrations.libs.const import PREDICTOR_STATUS
from mindsdb.interfaces.database.projects import ProjectController
from mindsdb.interfaces.jobs.jobs_controller import JobsController
from mindsdb.interfaces.chatbot.chatbot_controller import ChatBotController
from mindsdb.interfaces.storage.model_fs import HandlerStorage
from mindsdb.utilities.context import context as ctx
from mindsdb.utilities.functions import resolve_model_identifier
import mindsdb.utilities.profiler as profiler
from mindsdb.utilities.functions import mark_process


def _get_show_where(
        statement: ASTNode,
        from_name: Optional[str] = None,
        like_name: Optional[str] = None,
        initial: Optional[ASTNode] = None,
) -> ASTNode:
    """combine all possible show filters to single 'where' condition
    SHOW category [FROM name] [LIKE filter] [WHERE filter]

    Args:
        statement (ASTNode): 'show' query statement
        from_name (str): name of column for 'from' filter
        like_name (str): name of column for 'like' filter,
        initial (ASTNode): initial 'where' filter
    Returns:
        ASTNode: 'where' statemnt
    """
    where = []
    if initial is not None:
        where.append(initial)
    if statement.from_table is not None and from_name is not None:
        where.append(
            BinaryOperation(
                "=",
                args=[Identifier(from_name), Constant(statement.from_table.parts[-1])],
            )
        )
    if statement.like is not None and like_name is not None:
        where.append(
            BinaryOperation(
                "like", args=[Identifier(like_name), Constant(statement.like)]
            )
        )
    if statement.where is not None:
        where.append(statement.where)

    if len(where) > 0:
        return reduce(
            lambda prev, next: BinaryOperation("and", args=[prev, next]), where
        )
    return None


class ExecuteCommands:
    def __init__(self, session, executor):
        self.session = session
        self.executor = executor

        self.charset_text_type = CHARSET_NUMBERS["utf8_general_ci"]
        self.datahub = session.datahub

    @profiler.profile()
    def execute_command(self, statement):
        sql = None
        if self.executor is None:
            if isinstance(statement, ASTNode):
                sql = statement.to_string()
            sql_lower = sql.lower()
        else:
            sql = self.executor.sql
            sql_lower = self.executor.sql_lower

        if type(statement) == CreateDatabase:
            return self.answer_create_database(statement)
        elif type(statement) == CreateMLEngine:
            return self.answer_create_ml_engine(statement)
        elif type(statement) == DropMLEngine:
            return self.answer_drop_ml_engine(statement)
        elif type(statement) == DropPredictor:
            database_name = self.session.database
            if len(statement.name.parts) > 1:
                database_name = statement.name.parts[0].lower()
            model_name = statement.name.parts[-1]

            try:
                project = self.session.database_controller.get_project(database_name)
                project.drop_table(model_name)
            except Exception as e:
                if not statement.if_exists:
                    raise e
            return ExecuteAnswer(ANSWER_TYPE.OK)
        elif type(statement) == DropTables:
            return self.answer_drop_tables(statement)
        elif type(statement) == DropDatasource or type(statement) == DropDatabase:
            return self.answer_drop_database(statement)
        elif type(statement) == Describe:
            # NOTE in sql 'describe table' is same as 'show columns'
            return self.answer_describe_predictor(statement)
        elif type(statement) == RetrainPredictor:
            return self.answer_retrain_predictor(statement)
        elif type(statement) == FinetunePredictor:
            return self.answer_finetune_predictor(statement)
        elif type(statement) == Show:
            sql_category = statement.category.lower()
            if hasattr(statement, "modes"):
                if isinstance(statement.modes, list) is False:
                    statement.modes = []
                statement.modes = [x.upper() for x in statement.modes]
            if sql_category in ("predictors", "models"):
                where = BinaryOperation("=", args=[Constant(1), Constant(1)])

                new_statement = Select(
                    targets=[Star()],
                    from_table=Identifier(parts=["information_schema", "models"]),
                    where=_get_show_where(
                        statement, from_name="project", like_name="name"
                    ),
                )
                query = SQLQuery(new_statement, session=self.session)
                return self.answer_select(query)
            elif sql_category == "ml_engines":
                new_statement = Select(
                    targets=[Star()],
                    from_table=Identifier(parts=["information_schema", "ml_engines"]),
                    where=_get_show_where(statement, like_name="name"),
                )

                query = SQLQuery(new_statement, session=self.session)
                return self.answer_select(query)
            elif sql_category == "handlers":
                new_statement = Select(
                    targets=[Star()],
                    from_table=Identifier(parts=["information_schema", "handlers"]),
                    where=_get_show_where(statement, like_name="name"),
                )

                query = SQLQuery(new_statement, session=self.session)
                return self.answer_select(query)
            elif sql_category == "plugins":
                if statement.where is not None or statement.like:
                    raise SqlApiException(
                        "'SHOW PLUGINS' query should be used without filters"
                    )
                new_statement = Select(
                    targets=[Star()],
                    from_table=Identifier(parts=["information_schema", "PLUGINS"]),
                )
                query = SQLQuery(new_statement, session=self.session)
                return self.answer_select(query)
            elif sql_category in ("databases", "schemas"):
                new_statement = Select(
                    targets=[Identifier(parts=["NAME"], alias=Identifier("Database"))],
                    from_table=Identifier(parts=["information_schema", "DATABASES"]),
                    where=_get_show_where(statement, like_name="Database"),
                )

                if "FULL" in statement.modes:
                    new_statement.targets.extend(
                        [
                            Identifier(parts=["TYPE"], alias=Identifier("TYPE")),
                            Identifier(parts=["ENGINE"], alias=Identifier("ENGINE")),
                        ]
                    )

                query = SQLQuery(new_statement, session=self.session)
                return self.answer_select(query)
            elif sql_category in ("tables", "full tables"):
                schema = self.session.database or "mindsdb"
                if statement.from_table is not None:
                    schema = statement.from_table.parts[-1]
                    statement.from_table = None
                where = BinaryOperation(
                    "and",
                    args=[
                        BinaryOperation(
                            "=", args=[Identifier("table_schema"), Constant(schema)]
                        ),
                        BinaryOperation(
                            "or",
                            args=[
                                BinaryOperation(
                                    "=",
                                    args=[Identifier("table_type"), Constant("MODEL")],
                                ),
                                BinaryOperation(
                                    "or",
                                    args=[
                                        BinaryOperation(
                                            "=",
                                            args=[
                                                Identifier("table_type"),
                                                Constant("BASE TABLE"),
                                            ],
                                        ),
                                        BinaryOperation(
                                            "or",
                                            args=[
                                                BinaryOperation(
                                                    "=",
                                                    args=[
                                                        Identifier("table_type"),
                                                        Constant("SYSTEM VIEW"),
                                                    ],
                                                ),
                                                BinaryOperation(
                                                    "=",
                                                    args=[
                                                        Identifier("table_type"),
                                                        Constant("VIEW"),
                                                    ],
                                                ),
                                            ],
                                        ),
                                    ],
                                ),
                            ],
                        ),
                    ],
                )

                new_statement = Select(
                    targets=[
                        Identifier(
                            parts=["table_name"],
                            alias=Identifier(f"Tables_in_{schema}"),
                        )
                    ],
                    from_table=Identifier(parts=["information_schema", "TABLES"]),
                    where=_get_show_where(
                        statement, like_name=f"Tables_in_{schema}", initial=where
                    ),
                )

                if "FULL" in statement.modes:
                    new_statement.targets.append(
                        Identifier(parts=["TABLE_TYPE"], alias=Identifier("Table_type"))
                    )

                query = SQLQuery(new_statement, session=self.session)
                return self.answer_select(query)
            elif sql_category in (
                    "variables",
                    "session variables",
                    "session status",
                    "global variables",
            ):
                new_statement = Select(
                    targets=[
                        Identifier(parts=["Variable_name"]),
                        Identifier(parts=["Value"]),
                    ],
                    from_table=Identifier(parts=["dataframe"]),
                    where=_get_show_where(statement, like_name="Variable_name"),
                )

                data = {}
                is_session = "session" in sql_category
                for var_name, var_data in SERVER_VARIABLES.items():
                    var_name = var_name.replace("@@", "")
                    if is_session and var_name.startswith("session.") is False:
                        continue
                    if var_name.startswith("session.") or var_name.startswith(
                            "GLOBAL."
                    ):
                        name = var_name.replace("session.", "").replace("GLOBAL.", "")
                        data[name] = var_data[0]
                    elif var_name not in data:
                        data[var_name] = var_data[0]

                df = pd.DataFrame(data.items(), columns=["Variable_name", "Value"])
                data = query_df(df, new_statement)
                data = data.values.tolist()

                columns = [
                    Column(
                        name="Variable_name", table_name="session_variables", type="str"
                    ),
                    Column(name="Value", table_name="session_variables", type="str"),
                ]

                return ExecuteAnswer(
                    answer_type=ANSWER_TYPE.TABLE, columns=columns, data=data
                )
            elif sql_category == "search_path":
                return ExecuteAnswer(
                    answer_type=ANSWER_TYPE.TABLE,
                    columns=[
                        Column(
                            name="search_path", table_name="search_path", type="str"
                        )
                    ],
                    data=[["\"$user\", public"]]
                )
            elif "show status like 'ssl_version'" in sql_lower:
                return ExecuteAnswer(
                    answer_type=ANSWER_TYPE.TABLE,
                    columns=[
                        Column(
                            name="Value", table_name="session_variables", type="str"
                        ),
                        Column(
                            name="Value", table_name="session_variables", type="str"
                        ),
                    ],
                    data=[["Ssl_version", "TLSv1.1"]],
                )
            elif sql_category in ("function status", "procedure status"):
                # SHOW FUNCTION STATUS WHERE Db = 'MINDSDB';
                # SHOW PROCEDURE STATUS WHERE Db = 'MINDSDB'
                # SHOW FUNCTION STATUS WHERE Db = 'MINDSDB' AND Name LIKE '%';
                return self.answer_function_status()
            elif sql_category in ("index", "keys", "indexes"):
                # INDEX | INDEXES | KEYS are synonyms
                # https://dev.mysql.com/doc/refman/8.0/en/show-index.html
                new_statement = Select(
                    targets=[
                        Identifier("TABLE_NAME", alias=Identifier("Table")),
                        Identifier("NON_UNIQUE", alias=Identifier("Non_unique")),
                        Identifier("INDEX_NAME", alias=Identifier("Key_name")),
                        Identifier("SEQ_IN_INDEX", alias=Identifier("Seq_in_index")),
                        Identifier("COLUMN_NAME", alias=Identifier("Column_name")),
                        Identifier("COLLATION", alias=Identifier("Collation")),
                        Identifier("CARDINALITY", alias=Identifier("Cardinality")),
                        Identifier("SUB_PART", alias=Identifier("Sub_part")),
                        Identifier("PACKED", alias=Identifier("Packed")),
                        Identifier("NULLABLE", alias=Identifier("Null")),
                        Identifier("INDEX_TYPE", alias=Identifier("Index_type")),
                        Identifier("COMMENT", alias=Identifier("Comment")),
                        Identifier("INDEX_COMMENT", alias=Identifier("Index_comment")),
                        Identifier("IS_VISIBLE", alias=Identifier("Visible")),
                        Identifier("EXPRESSION", alias=Identifier("Expression")),
                    ],
                    from_table=Identifier(parts=["information_schema", "STATISTICS"]),
                    where=statement.where,
                )
                query = SQLQuery(new_statement, session=self.session)
                return self.answer_select(query)
            # FIXME if have answer on that request, then DataGrip show warning '[S0022] Column 'Non_unique' not found.'
            elif "show create table" in sql_lower:
                # SHOW CREATE TABLE `MINDSDB`.`predictors`
                table = sql[sql.rfind(".") + 1:].strip(" .;\n\t").replace("`", "")
                return self.answer_show_create_table(table)
            elif sql_category in ("character set", "charset"):
                new_statement = Select(
                    targets=[
                        Identifier("CHARACTER_SET_NAME", alias=Identifier("Charset")),
                        Identifier(
                            "DEFAULT_COLLATE_NAME", alias=Identifier("Description")
                        ),
                        Identifier(
                            "DESCRIPTION", alias=Identifier("Default collation")
                        ),
                        Identifier("MAXLEN", alias=Identifier("Maxlen")),
                    ],
                    from_table=Identifier(
                        parts=["INFORMATION_SCHEMA", "CHARACTER_SETS"]
                    ),
                    where=_get_show_where(statement, like_name="CHARACTER_SET_NAME"),
                )
                query = SQLQuery(new_statement, session=self.session)
                return self.answer_select(query)
            elif sql_category == "warnings":
                return self.answer_show_warnings()
            elif sql_category == "engines":
                new_statement = Select(
                    targets=[Star()],
                    from_table=Identifier(parts=["information_schema", "ENGINES"]),
                )
                query = SQLQuery(new_statement, session=self.session)
                return self.answer_select(query)
            elif sql_category == "collation":
                new_statement = Select(
                    targets=[
                        Identifier("COLLATION_NAME", alias=Identifier("Collation")),
                        Identifier("CHARACTER_SET_NAME", alias=Identifier("Charset")),
                        Identifier("ID", alias=Identifier("Id")),
                        Identifier("IS_DEFAULT", alias=Identifier("Default")),
                        Identifier("IS_COMPILED", alias=Identifier("Compiled")),
                        Identifier("SORTLEN", alias=Identifier("Sortlen")),
                        Identifier("PAD_ATTRIBUTE", alias=Identifier("Pad_attribute")),
                    ],
                    from_table=Identifier(parts=["INFORMATION_SCHEMA", "COLLATIONS"]),
                    where=_get_show_where(statement, like_name="Collation"),
                )
                query = SQLQuery(new_statement, session=self.session)
                return self.answer_select(query)
            elif sql_category == "table status":
                # TODO improve it
                # SHOW TABLE STATUS LIKE 'table'
                table_name = None
                if statement.like is not None:
                    table_name = statement.like
                # elif condition == 'from' and type(expression) == Identifier:
                #     table_name = expression.parts[-1]
                if table_name is None:
                    err_str = f"Can't determine table name in query: {sql}"
                    log.logger.warning(err_str)
                    raise ErTableExistError(err_str)
                return self.answer_show_table_status(table_name)
            elif sql_category == "columns":
                is_full = statement.modes is not None and "full" in statement.modes
                return self.answer_show_columns(
                    statement.from_table,
                    statement.where,
                    statement.like,
                    is_full=is_full,
                )
            else:
                raise ErNotSupportedYet(f"Statement not implemented: {sql}")
        elif type(statement) in (
                StartTransaction,
                CommitTransaction,
                RollbackTransaction,
        ):
            return ExecuteAnswer(ANSWER_TYPE.OK)
        elif type(statement) == Set:
            category = (statement.category or "").lower()
            if category == "" and type(statement.arg) == BinaryOperation:
                if statement.arg.args[0].parts[0].lower() == 'profiling':
                    if statement.arg.args[1].value in (1, True):
                        profiler.enable()
                        self.session.profiling = True
                    else:
                        profiler.disable()
                        self.session.profiling = False
                elif statement.arg.args[0].parts[0].lower() == 'predictor_cache':
                    if statement.arg.args[1].value in (1, True):
                        self.session.predictor_cache = True
                    else:
                        self.session.predictor_cache = False
                return ExecuteAnswer(ANSWER_TYPE.OK)
            elif category == "autocommit":
                return ExecuteAnswer(ANSWER_TYPE.OK)
            elif category == "names":
                # set names utf8;
                charsets = {
                    "utf8": CHARSET_NUMBERS["utf8_general_ci"],
                    "utf8mb4": CHARSET_NUMBERS["utf8mb4_general_ci"],
                }
                self.charset = statement.arg.parts[0]
                self.charset_text_type = charsets.get(self.charset)
                if self.charset_text_type is None:
                    log.logger.warning(
                        f"Unknown charset: {self.charset}. Setting up 'utf8_general_ci' as charset text type."
                    )
                    self.charset_text_type = CHARSET_NUMBERS["utf8_general_ci"]
                return ExecuteAnswer(
                    ANSWER_TYPE.OK,
                    state_track=[
                        ["character_set_client", self.charset],
                        ["character_set_connection", self.charset],
                        ["character_set_results", self.charset],
                    ],
                )
            else:
                log.logger.warning(
                    f"SQL statement is not processable, return OK package: {sql}"
                )
                return ExecuteAnswer(ANSWER_TYPE.OK)
        elif type(statement) == Use:
            db_name = statement.value.parts[-1]
            self.change_default_db(db_name)
            return ExecuteAnswer(ANSWER_TYPE.OK)
        elif type(statement) == CreatePredictor:
            return self.answer_create_predictor(statement)
        elif type(statement) == CreateView:
            return self.answer_create_view(statement)
        elif type(statement) == DropView:
            return self.answer_drop_view(statement)
        elif type(statement) == Delete:
            if statement.table.parts[-1].lower() == "models_versions":
                return self.answer_delete_model_version(statement)
            if (
                    self.session.database != "mindsdb"
                    and statement.table.parts[0] != "mindsdb"
            ):
                raise ErBadTableError(
                    "Only 'DELETE' from database 'mindsdb' is possible at this moment"
                )

            SQLQuery(statement, session=self.session, execute=True)
            return ExecuteAnswer(ANSWER_TYPE.OK)

        elif type(statement) == Insert:
            SQLQuery(statement, session=self.session, execute=True)
            return ExecuteAnswer(ANSWER_TYPE.OK)
        elif type(statement) == Update:
            if statement.from_select is None:
                if statement.table.parts[-1].lower() == "models_versions":
                    return self.answer_update_model_version(statement)

            SQLQuery(statement, session=self.session, execute=True)
            return ExecuteAnswer(ANSWER_TYPE.OK)
        elif (
                type(statement) == Alter
                and ("disable keys" in sql_lower)
                or ("enable keys" in sql_lower)
        ):
            return ExecuteAnswer(ANSWER_TYPE.OK)
        elif type(statement) == Select:
            if statement.from_table is None:
                return self.answer_single_row_select(statement)

            query = SQLQuery(statement, session=self.session)
            return self.answer_select(query)
        elif type(statement) == Union:
            query = SQLQuery(statement, session=self.session)
            return self.answer_select(query)
        elif type(statement) == Explain:
            return self.answer_show_columns(statement.target)
        elif type(statement) == CreateTable:
            # TODO
            return self.answer_apply_predictor(statement)
        # -- jobs --
        elif type(statement) == CreateJob:
            return self.answer_create_job(statement)
        elif type(statement) == DropJob:
            return self.answer_drop_job(statement)
        # -- chatbots --
        elif type(statement) == CreateChatBot:
            return self.answer_create_chatbot(statement)
        elif type(statement) == DropChatBot:
            return self.answer_drop_chatbot(statement)
        elif type(statement) == Evaluate:
            statement.data = parse_sql(statement.query_str, dialect='mindsdb')
            return self.answer_evaluate_metric(statement)
        else:
            log.logger.warning(f"Unknown SQL statement: {sql}")
            raise ErNotSupportedYet(f"Unknown SQL statement: {sql}")

    def answer_create_job(self, statement):
        jobs_controller = JobsController()

        name = statement.name
        job_name = name.parts[-1]
        project_name = name.parts[-2] if len(name.parts) > 1 else self.session.database

        jobs_controller.add(job_name, project_name, statement.query_str,
                            statement.start_str, statement.end_str, statement.repeat_str)

        return ExecuteAnswer(ANSWER_TYPE.OK)

    def answer_drop_job(self, statement):
        jobs_controller = JobsController()

        name = statement.name
        job_name = name.parts[-1]
        project_name = name.parts[-2] if len(name.parts) > 1 else self.session.database
        jobs_controller.delete(job_name, project_name)

        return ExecuteAnswer(ANSWER_TYPE.OK)

    def answer_create_chatbot(self, statement):
        chatbot_controller = ChatBotController()

        name = statement.name
        project_name = name.parts[-2] if len(name.parts) > 1 else self.session.database

        database = self.session.integration_controller.get(statement.database.parts[-1])
        if database is None:
            raise SqlApiException(f'Database not found: {statement.database}')

        chatbot_controller.add_chatbot(
            name.parts[-1],
            project_name=project_name,
            model_name=statement.model.parts[-1],
            database_id=database['id'],
            params=statement.params
        )
        return ExecuteAnswer(ANSWER_TYPE.OK)

    def answer_drop_chatbot(self, statement):
        chatbot_controller = ChatBotController()

        name = statement.name
        project_name = name.parts[-2] if len(name.parts) > 1 else self.session.database

        chatbot_controller.delete_chatbot(
            name.parts[-1],
            project_name=project_name
        )
        return ExecuteAnswer(ANSWER_TYPE.OK)

    def answer_evaluate_metric(self, statement):
        try:
            sqlquery = SQLQuery(statement.data, session=self.session)
        except Exception as e:
            raise Exception(f'Nested query failed to execute with error: "{e}", please check and try again.')
        result = sqlquery.fetch(self.session.datahub)
        df = pd.DataFrame.from_dict(result["result"])
        df.columns = [str(t.alias) if hasattr(t, 'alias') else str(t.parts[-1]) for t in statement.data.targets]

        for col in ['actual', 'prediction']:
            assert col in df.columns, f'`{col}` column was not provided, please try again.'
            assert df[col].isna().sum() == 0, f'There are missing values in the `{col}` column, please try again.'

        metric_name = statement.name.parts[-1]
        target_series = df.pop('prediction')
        using_clause = statement.using if statement.using is not None else {}
        metric_value = evaluate_accuracy(df, target_series, metric_name,
                                         target='actual',
                                         ts_analysis=using_clause.get('ts_analysis', {}),  # will be deprecated soon
                                         n_decimals=using_clause.get('n_decimals', 3))  # 3 decimals by default
        return ExecuteAnswer(
            answer_type=ANSWER_TYPE.TABLE,
            columns=[Column(name=metric_name, table_name='', type='str')],
            data=[[metric_value]]
        )

    def answer_describe_predictor(self, statement):

        # try full name
        attribute = None
        model_info = self._get_model_info(statement.value, except_absent=False)
        if model_info is None:
            parts = statement.value.parts.copy()
            attribute = parts.pop(-1)
            model_info = self._get_model_info(Identifier(parts=parts))
            if model_info is None:
                raise SqlApiException(f'Model not found: {statement.value}')

        df = self.session.model_controller.describe_model(
            self.session, model_info['project_name'], model_info['model_record'].name, attribute
        )

        df_dict = df.to_dict('split')

        columns = [
            Column(name=col, table_name='', type='str')
            for col in df_dict['columns']
        ]
        return ExecuteAnswer(
            answer_type=ANSWER_TYPE.TABLE,
            columns=columns,
            data=df_dict['data']
        )

    def _get_model_info(self, identifier, except_absent=True):
        if len(identifier.parts) == 1:
            identifier.parts = [self.session.database, identifier.parts[0]]

        database_name, model_name, model_version = resolve_model_identifier(identifier)

        if model_name is None:
            if except_absent:
                raise Exception(f'Model not found: {identifier.to_string()}')
            else:
                return

        model_record = get_model_record(
            name=model_name,
            project_name=database_name,
            except_absent=except_absent,
            version=model_version,
            active=True if model_version is None else None
        )
        if not model_record:
            return None
        return {
            'model_record': model_record,
            'project_name': database_name
        }

    def _sync_predictor_check(self, phase_name):
        """ Checks if there is already a predictor retraining or fine-tuning
            Do not allow to run retrain if there is another model in training process in less that 1h
        """
        is_cloud = self.session.config.get('cloud', False)
        if is_cloud and ctx.user_class == 0:
            models = get_model_records(active=None)
            shortest_training = None
            for model in models:
                if (
                        model.status in (PREDICTOR_STATUS.GENERATING, PREDICTOR_STATUS.TRAINING)
                        and model.training_start_at is not None and model.training_stop_at is None
                ):
                    training_time = datetime.datetime.now() - model.training_start_at
                    if shortest_training is None or training_time < shortest_training:
                        shortest_training = training_time

            if shortest_training is not None and shortest_training < datetime.timedelta(hours=1):
                raise SqlApiException(
                    f"Can't start {phase_name} process while predictor is in status 'training' or 'generating'"
                )

    def answer_retrain_predictor(self, statement):
        model_record = self._get_model_info(statement.name)['model_record']

        if statement.integration_name is None:
            if model_record.data_integration_ref is not None:
                if model_record.data_integration_ref["type"] == "integration":
                    integration = self.session.integration_controller.get_by_id(
                        model_record.data_integration_ref["id"]
                    )
                    if integration is None:
                        raise Exception(
                            "The database from which the model was trained no longer exists"
                        )

        ml_handler = None
        if statement.using is not None:
            # repack using with lower names
            statement.using = {k.lower(): v for k, v in statement.using.items()}

            if "engine" in statement.using:
                ml_integration_name = statement.using.pop("engine")
                ml_handler = self.session.integration_controller.get_handler(
                    ml_integration_name
                )

        # use current ml handler
        if ml_handler is None:
            integration_record = get_predictor_integration(model_record)
            if integration_record is None:
                raise Exception("ML engine model was trained with does not esxists")
            ml_handler = self.session.integration_controller.get_handler(
                integration_record.name
            )

        self._sync_predictor_check(phase_name='retrain')
        df = self.session.model_controller.retrain_model(statement, ml_handler)

        resp_dict = df.to_dict(orient='split')

        columns = [
            Column(col)
            for col in resp_dict['columns']
        ]

        return ExecuteAnswer(answer_type=ANSWER_TYPE.TABLE, columns=columns, data=resp_dict['data'])

    @profiler.profile()
    @mark_process('learn')
    def answer_finetune_predictor(self, statement):
        model_record = self._get_model_info(statement.name)['model_record']

        if statement.using is not None:
            # repack using with lower names
            statement.using = {k.lower(): v for k, v in statement.using.items()}

        # use current ml handler
        integration_record = get_predictor_integration(model_record)
        if integration_record is None:
            raise Exception('The ML engine that the model was trained with does not exist.')
        ml_handler = self.session.integration_controller.get_handler(integration_record.name)

        self._sync_predictor_check(phase_name='finetune')
        df = self.session.model_controller.finetune_model(statement, ml_handler)

        resp_dict = df.to_dict(orient='split')

        columns = [
            Column(col)
            for col in resp_dict['columns']
        ]

        return ExecuteAnswer(answer_type=ANSWER_TYPE.TABLE, columns=columns, data=resp_dict['data'])

    def _create_integration(self, name: str, engine: str, connection_args: dict):
        # we have connection checkers not for any db. So do nothing if fail
        # TODO return rich error message

        if connection_args is None:
            connection_args = {}
        status = HandlerStatusResponse(success=False)

        try:
            handlers_meta = (
                self.session.integration_controller.get_handlers_import_status()
            )
            handler_meta = handlers_meta[engine]
            if handler_meta.get("import", {}).get("success") is not True:
                raise SqlApiException(f"Handler '{engine}' can not be used")

            accept_connection_args = handler_meta.get("connection_args")
            if accept_connection_args is not None:
                for arg_name, arg_value in connection_args.items():
                    if arg_name == "as_service":
                        continue
                    if arg_name not in accept_connection_args:
                        raise SqlApiException(
                            f"Unknown connection argument: {arg_name}"
                        )
                    arg_meta = accept_connection_args[arg_name]
                    arg_type = arg_meta.get("type")
                    if arg_type == HANDLER_CONNECTION_ARG_TYPE.PATH:
                        # arg may be one of:
                        # str: '/home/file.pem'
                        # dict: {'path': '/home/file.pem'}
                        # dict: {'url': 'https://host.com/file'}
                        arg_value = connection_args[arg_name]
                        if isinstance(arg_value, (str, dict)) is False:
                            raise SqlApiException(f"Unknown type of arg: '{arg_value}'")
                        if isinstance(arg_value, str) or "path" in arg_value:
                            path = (
                                arg_value
                                if isinstance(arg_value, str)
                                else arg_value["path"]
                            )
                            if Path(path).is_file() is False:
                                raise SqlApiException(f"File not found at: '{path}'")
                        elif "url" in arg_value:
                            path = download_file(arg_value["url"])
                        else:
                            raise SqlApiException(
                                f"Argument '{arg_name}' must be path or url to the file"
                            )
                        connection_args[arg_name] = path

            handler = self.session.integration_controller.create_tmp_handler(
                handler_type=engine, connection_data=connection_args
            )
            status = handler.check_connection()
        except Exception as e:
            status.error_message = str(e)

        if status.success is False:
            raise SqlApiException(f"Can't connect to db: {status.error_message}")

        integration = self.session.integration_controller.get(name)
        if integration is not None:
            raise SqlApiException(f"Database '{name}' already exists.")

        self.session.integration_controller.add(name, engine, connection_args)

    def answer_create_ml_engine(self, statement: ASTNode):
        name = statement.name.parts[-1]
        integrations = self.session.integration_controller.get_all()
        if name in integrations:
            raise SqlApiException(f"Integration '{name}' already exists")

        handler_module_meta = (
            self.session.integration_controller.get_handlers_import_status().get(
                statement.handler
            )
        )
        if handler_module_meta is None:
            raise SqlApiException(f"There is no engine '{statement.handler}'")
        if handler_module_meta.get("import", {}).get("success") is not True:
            log.logger.info(f"to use {statement.handler} please install it 'pip install mindsdb[{statement.handler}]'")
            raise SqlApiException(f"Can't import engine '{statement.handler}'. to use it please install it 'pip install mindsdb[{statement.handler}]'")

        integration_id = self.session.integration_controller.add(
            name=name,
            engine=statement.handler,
            connection_args=statement.params
        )

        HandlerClass = self.session.integration_controller.handler_modules[handler_module_meta['name']].Handler

        if hasattr(HandlerClass, 'create_engine'):
            handlerStorage = HandlerStorage(integration_id)
            ml_handler = HandlerClass(
                engine_storage=handlerStorage,
                model_storage=None,
            )

            try:
                ml_handler.create_engine(statement.params)
            except NotImplementedError:
                pass
            except Exception as e:
                # something wrong, drop ml engine
                ast_drop = DropMLEngine(name=statement.name)
                self.answer_drop_ml_engine(ast_drop)
                raise e

        return ExecuteAnswer(ANSWER_TYPE.OK)

    def answer_drop_ml_engine(self, statement: ASTNode):
        name = statement.name.parts[-1]
        integrations = self.session.integration_controller.get_all()
        if name not in integrations:
            raise SqlApiException(f"Integration '{name}' does not exists")
        self.session.integration_controller.delete(name)
        return ExecuteAnswer(ANSWER_TYPE.OK)

    def answer_create_database(self, statement: ASTNode):
        """create new handler (datasource/integration in old terms)
        Args:
            statement (ASTNode): data for creating database/project
        """

        if len(statement.name.parts) != 1:
            raise Exception("Database name should contain only 1 part.")

        database_name = statement.name.parts[0]
        engine = statement.engine
        if engine is None:
            engine = "mindsdb"
        engine = engine.lower()
        connection_args = statement.parameters

        if engine == "mindsdb":
            ProjectController().add(database_name)
        else:
            self._create_integration(database_name, engine, connection_args)

        return ExecuteAnswer(ANSWER_TYPE.OK)

    def answer_drop_database(self, statement):
        if len(statement.name.parts) != 1:
            raise Exception("Database name should contain only 1 part.")
        db_name = statement.name.parts[0]
        self.session.database_controller.delete(db_name)
        return ExecuteAnswer(ANSWER_TYPE.OK)

    def answer_drop_tables(self, statement):
        """answer on 'drop table [if exists] {name}'
        Args:
            statement: ast
        """
        if statement.if_exists is False:
            for table in statement.tables:
                if len(table.parts) > 1:
                    db_name = table.parts[0]
                else:
                    db_name = self.session.database
                table_name = table.parts[-1]

                if db_name == "files":
                    dn = self.session.datahub[db_name]
                    if dn.has_table(table_name) is False:
                        raise SqlApiException(
                            f"Cannot delete a table from database '{db_name}': table does not exists"
                        )
                else:
                    projects_dict = self.session.database_controller.get_dict(
                        filter_type="project"
                    )
                    if db_name not in projects_dict:
                        raise SqlApiException(
                            f"Cannot delete a table from database '{db_name}'"
                        )
                    project = self.session.database_controller.get_project(db_name)
                    project_tables = {
                        key: val
                        for key, val in project.get_tables().items()
                        if val.get("deletable") is True
                    }
                    if table_name not in project_tables:
                        raise SqlApiException(
                            f"Cannot delete a table from database '{db_name}': table does not exists"
                        )

        for table in statement.tables:
            if len(table.parts) > 1:
                db_name = table.parts[0]
            else:
                db_name = self.session.database
            table_name = table.parts[-1]

            if db_name == "files":
                dn = self.session.datahub[db_name]
                if dn.has_table(table_name):
                    self.session.datahub["files"].query(
                        DropTables(tables=[Identifier(table_name)])
                    )
            else:
                projects_dict = self.session.database_controller.get_dict(
                    filter_type="project"
                )
                if db_name not in projects_dict:
                    continue
                self.session.model_controller.delete_model(
                    table_name, project_name=db_name
                )
        return ExecuteAnswer(ANSWER_TYPE.OK)

    def answer_create_view(self, statement):
        project_name = self.session.database
        # TEMP
        if isinstance(statement.name, Identifier):
            parts = statement.name.parts
        else:
            parts = statement.name.split(".")

        view_name = parts[-1]
        if len(parts) == 2:
            project_name = parts[0]

        query_str = statement.query_str
        query = parse_sql(query_str, dialect="mindsdb")

        if isinstance(statement.from_table, Identifier):
            query = Select(
                targets=[Star()],
                from_table=NativeQuery(
                    integration=statement.from_table, query=statement.query_str
                ),
            )
            renderer = SqlalchemyRender("mysql")
            query_str = renderer.get_string(query, with_failback=True)

        if isinstance(query, Select):
            # check create view sql
            query.limit = Constant(1)

            sqlquery = SQLQuery(query, session=self.session)
            if sqlquery.fetch()["success"] is not True:
                raise SqlApiException("Wrong view query")

        project = self.session.database_controller.get_project(project_name)
        project.create_view(view_name, query=query_str)
        return ExecuteAnswer(answer_type=ANSWER_TYPE.OK)

    def answer_drop_view(self, statement):
        names = statement.names

        for name in names:
            view_name = name.parts[-1]
            if len(name.parts) > 1:
                db_name = name.parts[0]
            else:
                db_name = self.session.database
            project = self.session.database_controller.get_project(db_name)
            project.drop_table(view_name)

        return ExecuteAnswer(answer_type=ANSWER_TYPE.OK)

    @mark_process('learn')
    def answer_create_predictor(self, statement):
        integration_name = self.session.database
        if len(statement.name.parts) > 1:
            integration_name = statement.name.parts[0]
        else:
            statement.name.parts = [integration_name, statement.name.parts[-1]]
        integration_name = integration_name.lower()

        ml_integration_name = "lightwood"
        if statement.using is not None:
            # repack using with lower names
            statement.using = {k.lower(): v for k, v in statement.using.items()}

            ml_integration_name = statement.using.pop("engine", ml_integration_name)

        ml_handler = self.session.integration_controller.get_handler(
            ml_integration_name
        )

        df = self.session.model_controller.create_model(statement, ml_handler)
        resp_dict = df.to_dict(orient='split')

        columns = [
            Column(col)
            for col in resp_dict['columns']
        ]

        return ExecuteAnswer(answer_type=ANSWER_TYPE.TABLE, columns=columns, data=resp_dict['data'])

    def answer_show_columns(
            self,
            target: Identifier,
            where: Optional[Operation] = None,
            like: Optional[str] = None,
            is_full=False,
    ):
        if len(target.parts) > 1:
            db = target.parts[0]
        elif isinstance(self.session.database, str) and len(self.session.database) > 0:
            db = self.session.database
        else:
            db = "mindsdb"
        table_name = target.parts[-1]

        new_where = BinaryOperation(
            "and",
            args=[
                BinaryOperation("=", args=[Identifier("TABLE_SCHEMA"), Constant(db)]),
                BinaryOperation(
                    "=", args=[Identifier("TABLE_NAME"), Constant(table_name)]
                ),
            ],
        )
        if where is not None:
            new_where = BinaryOperation("and", args=[new_where, where])
        if like is not None:
            like = BinaryOperation("like", args=[Identifier("View"), Constant(like)])
            new_where = BinaryOperation("and", args=[new_where, like])

        targets = [
            Identifier("COLUMN_NAME", alias=Identifier("Field")),
            Identifier("COLUMN_TYPE", alias=Identifier("Type")),
            Identifier("IS_NULLABLE", alias=Identifier("Null")),
            Identifier("COLUMN_KEY", alias=Identifier("Key")),
            Identifier("COLUMN_DEFAULT", alias=Identifier("Default")),
            Identifier("EXTRA", alias=Identifier("Extra")),
        ]
        if is_full:
            targets.extend(
                [
                    Constant("COLLATION", alias=Identifier("Collation")),
                    Constant("PRIVILEGES", alias=Identifier("Privileges")),
                    Constant("COMMENT", alias=Identifier("Comment")),
                ]
            )
        new_statement = Select(
            targets=targets,
            from_table=Identifier(parts=["information_schema", "COLUMNS"]),
            where=new_where,
        )

        query = SQLQuery(new_statement, session=self.session)
        return self.answer_select(query)

    def answer_single_row_select(self, statement):
        columns = []
        data = []
        for target in statement.targets:
            target_type = type(target)
            if target_type == Variable:
                var_name = target.value
                column_name = f"@@{var_name}"
                column_alias = target.alias or column_name
                result = SERVER_VARIABLES.get(column_name)
                if result is None:
                    log.logger.error(f"Unknown variable: {column_name}")
                    raise Exception(f"Unknown variable '{var_name}'")
                else:
                    result = result[0]
            elif target_type == Function:
                function_name = target.op.lower()
                if function_name == "connection_id":
                    return self.answer_connection_id()

                functions_results = {
                    # 'connection_id': self.executor.sqlserver.connection_id,
                    "database": self.session.database,
                    "current_user": self.session.username,
                    "user": self.session.username,
                    "version": "8.0.17",
                    "current_schema": "public"
                }

                column_name = f"{target.op}()"
                column_alias = target.alias or column_name
                result = functions_results[function_name]
            elif target_type == Constant:
                result = target.value
                column_name = str(result)
                column_alias = (
                    ".".join(target.alias.parts)
                    if type(target.alias) == Identifier
                    else column_name
                )
            elif target_type == NullConstant:
                result = None
                column_name = "NULL"
                column_alias = "NULL"
            elif target_type == Identifier:
                result = ".".join(target.parts)
                if result == "session_user":
                    column_name = result
                    result = self.session.username
                else:
                    raise Exception(f"Unknown column '{result}'")
            else:
                raise ErSqlWrongArguments(f"Unknown constant type: {target_type}")

            columns.append(
                Column(
                    name=column_name,
                    alias=column_alias,
                    table_name="",
                    type=TYPES.MYSQL_TYPE_VAR_STRING
                    if isinstance(result, str)
                    else TYPES.MYSQL_TYPE_LONG,
                    charset=self.charset_text_type
                    if isinstance(result, str)
                    else CHARSET_NUMBERS["binary"],
                )
            )
            data.append(result)

        return ExecuteAnswer(
            answer_type=ANSWER_TYPE.TABLE, columns=columns, data=[data]
        )

    def answer_show_create_table(self, table):
        columns = [
            Column(table_name="", name="Table", type=TYPES.MYSQL_TYPE_VAR_STRING),
            Column(
                table_name="", name="Create Table", type=TYPES.MYSQL_TYPE_VAR_STRING
            ),
        ]
        return ExecuteAnswer(
            answer_type=ANSWER_TYPE.TABLE,
            columns=columns,
            data=[[table, f"create table {table} ()"]],
        )

    def answer_function_status(self):
        columns = [
            Column(
                name="Db",
                alias="Db",
                table_name="schemata",
                table_alias="ROUTINES",
                type="str",
                database="mysql",
                charset=self.charset_text_type,
            ),
            Column(
                name="Db",
                alias="Db",
                table_name="routines",
                table_alias="ROUTINES",
                type="str",
                database="mysql",
                charset=self.charset_text_type,
            ),
            Column(
                name="Type",
                alias="Type",
                table_name="routines",
                table_alias="ROUTINES",
                type="str",
                database="mysql",
                charset=CHARSET_NUMBERS["utf8_bin"],
            ),
            Column(
                name="Definer",
                alias="Definer",
                table_name="routines",
                table_alias="ROUTINES",
                type="str",
                database="mysql",
                charset=CHARSET_NUMBERS["utf8_bin"],
            ),
            Column(
                name="Modified",
                alias="Modified",
                table_name="routines",
                table_alias="ROUTINES",
                type=TYPES.MYSQL_TYPE_TIMESTAMP,
                database="mysql",
                charset=CHARSET_NUMBERS["binary"],
            ),
            Column(
                name="Created",
                alias="Created",
                table_name="routines",
                table_alias="ROUTINES",
                type=TYPES.MYSQL_TYPE_TIMESTAMP,
                database="mysql",
                charset=CHARSET_NUMBERS["binary"],
            ),
            Column(
                name="Security_type",
                alias="Security_type",
                table_name="routines",
                table_alias="ROUTINES",
                type=TYPES.MYSQL_TYPE_STRING,
                database="mysql",
                charset=CHARSET_NUMBERS["utf8_bin"],
            ),
            Column(
                name="Comment",
                alias="Comment",
                table_name="routines",
                table_alias="ROUTINES",
                type=TYPES.MYSQL_TYPE_BLOB,
                database="mysql",
                charset=CHARSET_NUMBERS["utf8_bin"],
            ),
            Column(
                name="character_set_client",
                alias="character_set_client",
                table_name="character_sets",
                table_alias="ROUTINES",
                type=TYPES.MYSQL_TYPE_VAR_STRING,
                database="mysql",
                charset=self.charset_text_type,
            ),
            Column(
                name="collation_connection",
                alias="collation_connection",
                table_name="collations",
                table_alias="ROUTINES",
                type=TYPES.MYSQL_TYPE_VAR_STRING,
                database="mysql",
                charset=self.charset_text_type,
            ),
            Column(
                name="Database Collation",
                alias="Database Collation",
                table_name="collations",
                table_alias="ROUTINES",
                type=TYPES.MYSQL_TYPE_VAR_STRING,
                database="mysql",
                charset=self.charset_text_type,
            ),
        ]

        return ExecuteAnswer(answer_type=ANSWER_TYPE.TABLE, columns=columns, data=[])

    def answer_show_table_status(self, table_name):
        # NOTE at this moment parsed statement only like `SHOW TABLE STATUS LIKE 'table'`.
        # NOTE some columns has {'database': 'mysql'}, other not. That correct. This is how real DB sends messages.
        columns = [
            {
                "database": "mysql",
                "table_name": "tables",
                "name": "Name",
                "alias": "Name",
                "type": TYPES.MYSQL_TYPE_VAR_STRING,
                "charset": self.charset_text_type,
            },
            {
                "database": "",
                "table_name": "tables",
                "name": "Engine",
                "alias": "Engine",
                "type": TYPES.MYSQL_TYPE_VAR_STRING,
                "charset": self.charset_text_type,
            },
            {
                "database": "",
                "table_name": "tables",
                "name": "Version",
                "alias": "Version",
                "type": TYPES.MYSQL_TYPE_LONGLONG,
                "charset": CHARSET_NUMBERS["binary"],
            },
            {
                "database": "mysql",
                "table_name": "tables",
                "name": "Row_format",
                "alias": "Row_format",
                "type": TYPES.MYSQL_TYPE_VAR_STRING,
                "charset": self.charset_text_type,
            },
            {
                "database": "",
                "table_name": "tables",
                "name": "Rows",
                "alias": "Rows",
                "type": TYPES.MYSQL_TYPE_LONGLONG,
                "charset": CHARSET_NUMBERS["binary"],
            },
            {
                "database": "",
                "table_name": "tables",
                "name": "Avg_row_length",
                "alias": "Avg_row_length",
                "type": TYPES.MYSQL_TYPE_LONGLONG,
                "charset": CHARSET_NUMBERS["binary"],
            },
            {
                "database": "",
                "table_name": "tables",
                "name": "Data_length",
                "alias": "Data_length",
                "type": TYPES.MYSQL_TYPE_LONGLONG,
                "charset": CHARSET_NUMBERS["binary"],
            },
            {
                "database": "",
                "table_name": "tables",
                "name": "Max_data_length",
                "alias": "Max_data_length",
                "type": TYPES.MYSQL_TYPE_LONGLONG,
                "charset": CHARSET_NUMBERS["binary"],
            },
            {
                "database": "",
                "table_name": "tables",
                "name": "Index_length",
                "alias": "Index_length",
                "type": TYPES.MYSQL_TYPE_LONGLONG,
                "charset": CHARSET_NUMBERS["binary"],
            },
            {
                "database": "",
                "table_name": "tables",
                "name": "Data_free",
                "alias": "Data_free",
                "type": TYPES.MYSQL_TYPE_LONGLONG,
                "charset": CHARSET_NUMBERS["binary"],
            },
            {
                "database": "",
                "table_name": "tables",
                "name": "Auto_increment",
                "alias": "Auto_increment",
                "type": TYPES.MYSQL_TYPE_LONGLONG,
                "charset": CHARSET_NUMBERS["binary"],
            },
            {
                "database": "",
                "table_name": "tables",
                "name": "Create_time",
                "alias": "Create_time",
                "type": TYPES.MYSQL_TYPE_TIMESTAMP,
                "charset": CHARSET_NUMBERS["binary"],
            },
            {
                "database": "",
                "table_name": "tables",
                "name": "Update_time",
                "alias": "Update_time",
                "type": TYPES.MYSQL_TYPE_TIMESTAMP,
                "charset": CHARSET_NUMBERS["binary"],
            },
            {
                "database": "",
                "table_name": "tables",
                "name": "Check_time",
                "alias": "Check_time",
                "type": TYPES.MYSQL_TYPE_TIMESTAMP,
                "charset": CHARSET_NUMBERS["binary"],
            },
            {
                "database": "mysql",
                "table_name": "tables",
                "name": "Collation",
                "alias": "Collation",
                "type": TYPES.MYSQL_TYPE_VAR_STRING,
                "charset": self.charset_text_type,
            },
            {
                "database": "",
                "table_name": "tables",
                "name": "Checksum",
                "alias": "Checksum",
                "type": TYPES.MYSQL_TYPE_LONGLONG,
                "charset": CHARSET_NUMBERS["binary"],
            },
            {
                "database": "",
                "table_name": "tables",
                "name": "Create_options",
                "alias": "Create_options",
                "type": TYPES.MYSQL_TYPE_VAR_STRING,
                "charset": self.charset_text_type,
            },
            {
                "database": "",
                "table_name": "tables",
                "name": "Comment",
                "alias": "Comment",
                "type": TYPES.MYSQL_TYPE_BLOB,
                "charset": self.charset_text_type,
            },
        ]
        columns = [Column(**d) for d in columns]
        data = [
            [
                table_name,  # Name
                "InnoDB",  # Engine
                10,  # Version
                "Dynamic",  # Row_format
                1,  # Rows
                16384,  # Avg_row_length
                16384,  # Data_length
                0,  # Max_data_length
                0,  # Index_length
                0,  # Data_free
                None,  # Auto_increment
                datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"),  # Create_time
                datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"),  # Update_time
                None,  # Check_time
                "utf8mb4_0900_ai_ci",  # Collation
                None,  # Checksum
                "",  # Create_options
                "",  # Comment
            ]
        ]
        return ExecuteAnswer(answer_type=ANSWER_TYPE.TABLE, columns=columns, data=data)

    def answer_show_warnings(self):
        columns = [
            {
                "database": "",
                "table_name": "",
                "name": "Level",
                "alias": "Level",
                "type": TYPES.MYSQL_TYPE_VAR_STRING,
                "charset": self.charset_text_type,
            },
            {
                "database": "",
                "table_name": "",
                "name": "Code",
                "alias": "Code",
                "type": TYPES.MYSQL_TYPE_LONG,
                "charset": CHARSET_NUMBERS["binary"],
            },
            {
                "database": "",
                "table_name": "",
                "name": "Message",
                "alias": "Message",
                "type": TYPES.MYSQL_TYPE_VAR_STRING,
                "charset": self.charset_text_type,
            },
        ]
        columns = [Column(**d) for d in columns]
        return ExecuteAnswer(answer_type=ANSWER_TYPE.TABLE, columns=columns, data=[])

    def answer_connection_id(self):
        columns = [
            {
                "database": "",
                "table_name": "",
                "name": "conn_id",
                "alias": "conn_id",
                "type": TYPES.MYSQL_TYPE_LONG,
                "charset": CHARSET_NUMBERS["binary"],
            }
        ]
        columns = [Column(**d) for d in columns]
        data = [[self.executor.sqlserver.connection_id]]
        return ExecuteAnswer(answer_type=ANSWER_TYPE.TABLE, columns=columns, data=data)

    def answer_apply_predictor(self, statement):
        SQLQuery(statement, session=self.session, execute=True)
        return ExecuteAnswer(ANSWER_TYPE.OK)

    def answer_select(self, query):
        data = query.fetch()

        return ExecuteAnswer(
            answer_type=ANSWER_TYPE.TABLE,
            columns=query.columns_list,
            data=data["result"],
        )

    def answer_update_model_version(self, statement):

        # get project name
        if len(statement.table.parts) > 1:
            project_name = statement.table.parts[0]
        else:
            project_name = self.session.database

        project_datanode = self.datahub.get(project_name)
        if project_datanode is None:
            raise Exception(f"Project not found: {project_name}")

        # get list of model versions using filter
        query = Select(
            targets=[Identifier("version"), Identifier("name"), Identifier("project")],
            from_table=Identifier("models_versions"),
            where=statement.where,
        )

        models, _ = project_datanode.query(query=query, session=self.session)

        # get columns for update
        kwargs = {}
        for k, v in statement.update_columns.items():
            if isinstance(v, Constant):
                v = v.value
            kwargs[k.lower()] = v
        self.session.model_controller.update_model_version(models, **kwargs)
        return ExecuteAnswer(ANSWER_TYPE.OK)

    def answer_delete_model_version(self, statement):
        # get project name
        if len(statement.table.parts) > 1:
            project_name = statement.table.parts[0]
        else:
            project_name = self.session.database

        project_datanode = self.datahub.get(project_name)
        if project_datanode is None:
            raise Exception(f"Project not found: {project_name}")

        # get list of model versions using filter
        query = Select(
            targets=[Identifier("version"), Identifier("name"), Identifier("project")],
            from_table=Identifier("models_versions"),
            where=statement.where,
        )

        models, _ = project_datanode.query(query=query, session=self.session)

        self.session.model_controller.delete_model_version(models)
        return ExecuteAnswer(ANSWER_TYPE.OK)

    def change_default_db(self, db_name):
        # That fix for bug in mssql: it keeps connection for a long time, but after some time mssql can
        # send packet with COM_INIT_DB=null. In this case keep old database name as default.
        if db_name != "null":
            if self.session.database_controller.exists(db_name):
                self.session.database = db_name
            else:
                raise ErBadDbError(f"Database {db_name} does not exists")



from setuptools import setup, find_packages
from mindsdb.integrations.handlers.autosklearn_handler.__about__ import __title__, __version__, __github__, __pypi__, __license__, __author__, __description__

with open('requirements.txt') as req_file:
    requirements = [req.strip() for req in req_file.read().splitlines()]

setup(
    name=__title__,
    version=__version__,
    url=__github__,
    download_url=__pypi__,
    license=__license__,
    author=__author__,
    description=__description__,
    packages=find_packages(),
    install_requires=requirements,
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.8"
)



from mindsdb.integrations.handlers.confluence_handler.confluence_table import ConfluenceSpacesTable
from mindsdb.integrations.libs.api_handler import APIHandler
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
)
from mindsdb.utilities.log import get_log
from mindsdb_sql import parse_sql
from mindsdb.utilities.log import get_log

from atlassian import Confluence
from typing import Optional
import requests

logger = get_log("integrations.confluence_handler")

class ConfluenceHandler(APIHandler):
    """Confluence handler implementation"""

    def __init__(self, name=None, **kwargs):
        """Initialize the Confluence handler.
        Parameters
        ----------
        name : str
            name of a handler instance
        """
        super().__init__(name)
        
        connection_data = kwargs.get("connection_data", {})
        
        self.parser = parse_sql
        self.dialect = 'confluence'
        self.connection_data = connection_data
        self.kwargs = kwargs
        self.connection = None
        self.is_connected = False

        confluence_pages_data = ConfluenceSpacesTable(self)
        self._register_table("pages", confluence_pages_data)

    def connect(self) -> StatusResponse:
        """Set up the connection required by the handler.
        Returns
        -------
        StatusResponse
            connection object
        """
        if self.is_connected is True:
            return self.connection
       
        s = requests.Session()
        s.headers['Authorization'] =  f"Bearer {self.connection_data['confluence_api_token']}"

        self.connection = Confluence(url= self.connection_data['url'], session=s)
        self.is_connected = True
        
        return self.connection

    def check_connection(self) -> StatusResponse:
        """Check connection to the handler.
        Returns
        -------
        StatusResponse
            Status confirmation
        """
        response = StatusResponse(False)
        need_to_close = self.is_connected is False
        
        try:
            self.connect()
            response.success = True
        except Exception as e:
            logger.error(f"Error connecting to Confluence API: {e}!")
            response.error_message = e

        self.is_connected = response.success

        return response

    def native_query(self, query: str) -> StatusResponse:
        """Receive and process a raw query.
        Parameters
        ----------
        query : str
            query in a native format
        Returns
        -------
        StatusResponse
            Request status
        """
        ast = parse_sql(query, dialect="mindsdb")
        return self.query(ast)



from slack_sdk.web import WebClient
from slack_sdk.errors import SlackApiError
from slack_sdk.socket_mode import SocketModeClient
from slack_sdk.socket_mode.request import SocketModeRequest
from slack_sdk.socket_mode.response import SocketModeResponse

from threading import Event
from typing import Callable, Dict

from mindsdb.integrations.libs.realtime_chat_handler import RealtimeChatHandler
from mindsdb.interfaces.chatbot.chatbot_message import ChatBotMessage
from mindsdb.interfaces.chatbot.chatbot_response import ChatBotResponse

class RealtimeSlackChatHandler(RealtimeChatHandler):
    """Implements RealtimeChatHandler interface for sending/receiving Slack messages."""

    def __init__(self, on_message: Callable[[ChatBotMessage], None], params: Dict[str, str]):
        super().__init__('SlackChatHandler', on_message)

        if 'app_token' not in params or 'web_token' not in params:
            raise ValueError('Need app_token and web_token parameters to use Slack chat handler')
        self._stop_event = Event()
        self._socket_mode_client = SocketModeClient(
            # This app-level token will be used only for establishing a connection
            app_token=params['app_token'],  # xapp-A111-222-xyz
            # You will be using this WebClient for performing Web API calls in listeners
            web_client=WebClient(token=params['web_token'])  # xoxb-111-222-xyz
        )
        self._socket_mode_client.socket_mode_request_listeners.append(self._process_websocket_message)

    def _process_websocket_message(self, client: SocketModeClient, request: SocketModeRequest):
        # Acknowledge the request
        response = SocketModeResponse(envelope_id=request.envelope_id)
        client.send_socket_mode_response(response)

        if request.type != 'events_api':
            return
        
        payload_event = request.payload['event']
        if payload_event['type'] != 'message':
            return
        if 'subtype' in payload_event:
            # Don't respond to message_changed, message_deleted, etc.
            return
        if payload_event['channel_type'] != 'im':
            # Only support IMs currently.
            return
        if 'bot_id' in payload_event:
            # A bot sent this message.
            return

        chatbot_message = ChatBotMessage(
            ChatBotMessage.Type.DIRECT,
            payload_event['text'],
            # In Slack direct messages are treated as channels themselves.
            payload_event['channel'],
            payload_event['channel']
        )
        self.on_message(chatbot_message)

    def connect(self):
        """Begins listening for Slack messages using underlying websocket client."""
        self._socket_mode_client.connect()
        self._stop_event.wait()
    
    def disconnect(self):
        """Stops listening for Slack messages."""
        self._stop_event.set()
    
    def send_message(self, message: ChatBotMessage) -> ChatBotResponse:
        """
        Sends a Slack message.
        
        Parameters: message (ChatBotMessage): The message to send

        Returns: response (ChatBotResponse): Response indicating whether the message was sent successfully
        """
        if message.type != ChatBotMessage.Type.DIRECT:
            raise NotImplementedError('Only sending direct messages is supported by RealtimeSlackChatHandler')
        response = self._socket_mode_client.web_client.chat_postMessage(
            channel=message.destination,
            text=message.text
        )
        try:
            response.validate()
            return ChatBotResponse(message.text)
        except SlackApiError as e:
            return ChatBotResponse(message.text, error=str(e))

        


"""view_integration

Revision ID: 6a54ba55872e
Revises: 6e834843e7e9
Create Date: 2022-08-25 14:52:44.269082

"""
from alembic import op
import sqlalchemy as sa
import mindsdb.interfaces.storage.db    # noqa


# revision identifiers, used by Alembic.
revision = '6a54ba55872e'
down_revision = '976f15a37e6a'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###

    with op.batch_alter_table('view', schema=None) as batch_op:
        batch_op.drop_constraint('fk_integration_id', type_='foreignkey')
        batch_op.drop_column('integration_id')

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('view', schema=None) as batch_op:
        batch_op.add_column(sa.Column('integration_id', sa.INTEGER(), autoincrement=False, nullable=False))
        batch_op.create_foreign_key('fk_integration_id', 'integration', ['integration_id'], ['id'])

    # ### end Alembic commands ###



from mindsdb.api.mongo.classes import Responder
import mindsdb.api.mongo.functions as helpers
from mindsdb_sql.parser.ast import Delete, Identifier, BinaryOperation, Constant
from mindsdb_sql.parser.dialects.mindsdb import DropPredictor, DropJob, DropMLEngine
from mindsdb.interfaces.jobs.jobs_controller import JobsController

from mindsdb.api.mongo.classes.query_sql import run_sql_command


class Responce(Responder):
    when = {'delete': helpers.is_true}

    def result(self, query, request_env, mindsdb_env, session):
        try:
            res = self._result(query, request_env, mindsdb_env)
        except Exception as e:
            res = {
                'n': 0,
                'writeErrors': [{
                    'index': 0,
                    'code': 0,
                    'errmsg': str(e)
                }],
                'ok': 1
            }
        return res

    def _result(self, query, request_env, mindsdb_env):
        table = query['delete']
        if table == 'predictors':
            table = 'models'

        project_name = request_env['database']

        allowed_tables = ('models_versions', 'models', 'jobs', 'ml_engines')
        if table not in allowed_tables:
            raise Exception(f"Only removing from this collections is supported: {', '.join(allowed_tables)}")

        if len(query['deletes']) != 1:
            raise Exception("Should be only one argument in REMOVE operation")

        obj_name, obj_id = None, None

        delete_filter = query['deletes'][0]['q']
        if '_id' in delete_filter:
            # get name of object
            obj_id = helpers.objectid_to_int(delete_filter['_id'])

        if 'name' in delete_filter:
            obj_name = delete_filter['name']

        version = None
        if 'version' in delete_filter:
            version = delete_filter['version']

        if obj_name is None and obj_id is None:
            raise Exception("Can't find object to delete, use filter by name or _id")

        if obj_name is None:
            if table == 'models' or table == 'models_versions':
                model_id = obj_id >> 20
                if obj_name is None:
                    models = mindsdb_env['model_controller'].get_models(
                        ml_handler_name=None,
                        project_name=project_name
                    )
                    for model in models:
                        if model['id'] == model_id:
                            obj_name = model['name']
                            break
                    if obj_name is None:
                        raise Exception("Can't find model by _id")
            elif table == 'jobs':
                jobs_controller = JobsController()
                for job in jobs_controller.get_list(project_name):
                    if job['id'] == obj_id:
                        obj_name = job['name']
                        break

        # delete model
        if table == 'models':
            ast_query = DropPredictor(Identifier(parts=[project_name, obj_name]))
            run_sql_command(request_env, ast_query)

        # delete model version
        elif table == 'models_versions':
            if version is None:
                if obj_id is None:
                    raise Exception("Can't find object version")

                version = obj_id & (2**20 - 1)

            ast_query = Delete(table=Identifier(parts=[project_name, 'models_versions']),
                               where=BinaryOperation(op='and', args=[
                                   BinaryOperation(op='=', args=[Identifier('name'), Constant(obj_name)]),
                                   BinaryOperation(op='=', args=[Identifier('version'), Constant(version)])
                               ]))
            run_sql_command(request_env, ast_query)

        elif table == 'jobs':
            ast_query = DropJob(Identifier(parts=[project_name, obj_name]))
            run_sql_command(request_env, ast_query)

        elif table == 'ml_engines':
            ast_query = DropMLEngine(Identifier(parts=[obj_name]))
            run_sql_command(request_env, ast_query)

        return {
            'n': 1,
            'ok': 1
        }


responder = Responce()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ml.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


import mindsdb.microservices_grpc.ml.common_pb2 as common__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x08ml.proto\x1a\x0c\x63ommon.proto\"i\n\x10HandlerContextML\x12\x16\n\x0eintegration_id\x18\x01 \x01(\x05\x12\x14\n\x0cpredictor_id\x18\x02 \x01(\x05\x12\x0f\n\x07\x63ontext\x18\x03 \x01(\t\x12\x16\n\x0ehandler_params\x18\x04 \x01(\t\"K\n\x0bPredictCall\x12\"\n\x07\x63ontext\x18\x01 \x01(\x0b\x32\x11.HandlerContextML\x12\n\n\x02\x64\x66\x18\x02 \x01(\x0c\x12\x0c\n\x04\x61rgs\x18\x03 \x01(\t\"Z\n\nCreateCall\x12\"\n\x07\x63ontext\x18\x01 \x01(\x0b\x32\x11.HandlerContextML\x12\x0e\n\x06target\x18\x02 \x01(\t\x12\n\n\x02\x64\x66\x18\x03 \x01(\x0c\x12\x0c\n\x04\x61rgs\x18\x04 \x01(\t\"J\n\nUpdateCall\x12\"\n\x07\x63ontext\x18\x01 \x01(\x0b\x32\x11.HandlerContextML\x12\n\n\x02\x64\x66\x18\x03 \x01(\x0c\x12\x0c\n\x04\x61rgs\x18\x04 \x01(\t2\x85\x01\n\tMLService\x12$\n\x07Predict\x12\x0c.PredictCall\x1a\t.Response\"\x00\x12(\n\x06\x43reate\x12\x0b.CreateCall\x1a\x0f.StatusResponse\"\x00\x12(\n\x06Update\x12\x0b.UpdateCall\x1a\x0f.StatusResponse\"\x00\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'ml_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _HANDLERCONTEXTML._serialized_start=26
  _HANDLERCONTEXTML._serialized_end=131
  _PREDICTCALL._serialized_start=133
  _PREDICTCALL._serialized_end=208
  _CREATECALL._serialized_start=210
  _CREATECALL._serialized_end=300
  _UPDATECALL._serialized_start=302
  _UPDATECALL._serialized_end=376
  _MLSERVICE._serialized_start=379
  _MLSERVICE._serialized_end=512
# @@protoc_insertion_point(module_scope)



class ChatBotResponse:
    """
    Represents a response from a chatbot sending a message over a chat application.

    Attributes:
        text (str): Text of the message sent by the chatbot
        error (str): If sending the message failed, the cause of failure
    """
    def __init__(self, text: str, error: str = None):
        self.text = text
        self.error = error
